#include <boost/bind.hpp>
#include <map>
#include <vector>

#include "building.hpp"
#include "button.hpp"
#include "deity.hpp"
#include "foreach.hpp"
#include "formatter.hpp"
#include "game.hpp"
#include "game_dialog.hpp"
#include "grid_widget.hpp"
#include "image_widget.hpp"
#include "label.hpp"
#include "player.hpp"
#include "random.hpp"
#include "raster.hpp"

namespace {
std::vector<building_ptr> all_buildings;
std::map<std::string, int> building_ids;

int add_building(building_ptr b)
{
	b->set_id(all_buildings.size());
	building_ids[b->name()] = all_buildings.size();
	all_buildings.push_back(b);
	return all_buildings.size();
}

#define BUILDING_DEF(name, cost1, cost2, cost3, cost4, culture, points, unique, help) \
  class building_##name : public building { \
  public: \
    building_##name() : building(#name, cost1,cost2,cost3,cost4, culture, points, unique, help) { \
	} \
	int ai_evaluate(const player& p) const; \
  private: \
	void on_build(player& p) const; \
  }; \
  const int name##_building_id = add_building(building_ptr(new building_##name)); \
void building_##name::on_build(player& p) const

#define BUILDING_AI(name) int building_##name::ai_evaluate(const player& p) const

//           BUILDING    GRAIN  VENISON  FISH  ANY  CULTURE  VPs  UNIQUE
BUILDING_DEF(agora,          0,       0,    0,   2,       1,   0, false,
   "Allows population to grow beyond 5") {
	p.set_max_population_level(9);
}

BUILDING_AI(agora) {
	const int headroom = 4 - p.development(DEVELOP_POPULATION);
	switch(headroom) {
	case 0: return 600;
	case 1: return 400;
	default: return 200;
	}
}

BUILDING_DEF(barracks,       0,       1,    1,   0,       1,   1, false,
   "Allows military to grow beyond 3") {
	p.set_max_military_level(5);
	if(p.has_building(get_building_id("army"))) {
		p.set_max_military_level(7);
	}
}

BUILDING_AI(barracks) {
	const int headroom = 2 - p.development(DEVELOP_MILITARY);
	switch(headroom) {
	case 0: return 500;
	case 1: return 300;
	default: return 200;
	}
}

BUILDING_DEF(gymnasium,      1,       1,    0,   0,       1,   1, false,
   "Buildings require one less point of culture") {
	p.set_building_culture_reduction(1);
}

BUILDING_AI(gymnasium) { return 200; }

BUILDING_DEF(well,           2,       0,    0,   0,       1,   1, false,
   "+1 grain development") {
	p.develop(DEVELOP_GRAIN, 1);
}
BUILDING_AI(well) { return 100 + p.ai_rate_development(DEVELOP_GRAIN, 1); }

BUILDING_DEF(gardens,        0,       2,    0,   0,       1,   1, false,
   "+1 venison development") {
	p.develop(DEVELOP_VENISON, 1);
}
BUILDING_AI(gardens) { return 100 + p.ai_rate_development(DEVELOP_VENISON, 1); }

BUILDING_DEF(harbor,         0,       0,    2,   0,       1,   1, false,
   "+1 fish development") {
	p.develop(DEVELOP_FISH, 1);
}
BUILDING_AI(harbor) { return 100 + p.ai_rate_development(DEVELOP_FISH, 1); }

BUILDING_DEF(road_system,    1,       0,    1,   0,       1,   1, false,
   "+1 population") {
	p.develop(DEVELOP_POPULATION, 1);
}
BUILDING_AI(road_system) { return 100 + p.ai_rate_development(DEVELOP_POPULATION, 1); }

BUILDING_DEF(temple_athena,  1,       0,    2,   0,       2,   1, true,
   "Worshipping Athena gives a free development.\nIf culture > 5, each opponent regresses a development when you worship Athena.") {
	p.set_temple(deity::get_by_name("athena")->number());
}
BUILDING_AI(temple_athena) { return 800; }

BUILDING_DEF(temple_aphrodites,  2,       1,    0,   0,       2,   1, true,
    "Worshipping Aphrodite gives 2 free resources\nplus an additional resource if you have 10 population.") {
	p.set_temple(deity::get_by_name("aphrodite")->number());
}
BUILDING_AI(temple_aphrodites) { return 800; }

BUILDING_DEF(temple_ares,    0,       2,    1,   0,       2,   1, true,
    "When you win a war, a priest is restored to you.") {
	p.set_temple(deity::get_by_name("ares")->number());
	p.set_gets_priest_when_win_war();
}
BUILDING_AI(temple_ares) { return 600 + p.development(DEVELOP_MILITARY)*100; }

BUILDING_DEF(temple_hephaestus, 0,       0,    0,   4,       2,   1, true,
    "Once per building, you may pay a VP\nin place of a resource") {
	p.set_temple(deity::get_by_name("hephaestus")->number());
	p.allow_victory_points_payment_for_building(1);
}
BUILDING_AI(temple_hephaestus) { return 700; }
BUILDING_DEF(statue,         2,       1,    1,   0,       3,   3, false,
    "+1 culture, +1 food development of your choice") {
	p.develop(DEVELOP_CULTURE, 1);
	if(p.is_local_ai()) {
		p.develop(static_cast<DEVELOPMENT_TYPE>(DEVELOP_GRAIN + rng::generate()%3), 1);
	} else {
		using namespace gui;
		game_dialog& dia = game_dialog::get();
		dialog* d = dia.set_message(&p, "Choose which area to develop");
		grid_ptr g(new grid(1));

		for(int n = 0; n != 3; ++n) {
			const DEVELOPMENT_TYPE type = static_cast<DEVELOPMENT_TYPE>(DEVELOP_GRAIN + n);
			g->add_col(widget_ptr(new image_widget(player::development_icon(type))));
		}

		g->allow_selection();
		g->register_selection_callback(boost::bind(&game_dialog::make_selection, &dia, _1));
		d->add_widget(g);

		const int nitem = dia.modal_selection();
		p.develop(static_cast<DEVELOPMENT_TYPE>(DEVELOP_GRAIN + nitem), 1);
		dia.clear_message();
	}
}
BUILDING_AI(statue) { return p.ai_rate_development(DEVELOP_CULTURE) + std::max(std::max(p.ai_rate_development(DEVELOP_GRAIN), p.ai_rate_development(DEVELOP_FISH)), p.ai_rate_development(DEVELOP_VENISON)) + p.ai_rate_victory_points(victory_points()); }
BUILDING_DEF(lyceum,         1,       2,    1,   0,       3,   3, false,
  "+1 population, +1 culture") {
	p.develop(DEVELOP_CULTURE, 1);
	p.develop(DEVELOP_POPULATION, 1);
}
BUILDING_AI(lyceum) { return p.ai_rate_development(DEVELOP_CULTURE) + p.ai_rate_development(DEVELOP_POPULATION) + p.ai_rate_victory_points(victory_points()); }
BUILDING_DEF(militia,        1,       1,    2,   0,       3,   3, false,
  "+1 military, +1 population") {
	p.develop(DEVELOP_MILITARY, 1);
	p.develop(DEVELOP_POPULATION, 1);
}
BUILDING_AI(militia) { return p.ai_rate_development(DEVELOP_MILITARY) + p.ai_rate_development(DEVELOP_POPULATION) + p.ai_rate_victory_points(victory_points()); }
BUILDING_DEF(workshop,       0,       1,    1,   0,       3,   1, false,
  "+1 military or +1 food development or production or +1 VP\nThe first to build the workshop gets two effects") {
	//if nobody else has the workshop we get two benefits instead of one.
	int nchoices = 2;
	for(int n = 0; n != p.get_game().players().size(); ++n) {
		if(n == p.number()) {
			continue;
		}

		if(p.get_game().players()[n]->has_building(id())) {
			nchoices = 1;
			break;
		}
	}

	const DEVELOPMENT_TYPE devel_types[] = {DEVELOP_GRAIN, DEVELOP_VENISON, DEVELOP_FISH, DEVELOP_MILITARY};
	int effect_used = -1;
	while(nchoices--) {
		if(p.is_local_ai()) {
			p.produce(static_cast<RESOURCE_TYPE>(nchoices));
			continue;
		}

		using namespace gui;
		game_dialog& dia = game_dialog::get();
		dialog* d = dia.set_message(&p, "Choose what to build in the Workshop");
		grid_ptr g(new grid(2));

		int neffect = 0;

		if(effect_used != neffect++) {
			g->add_col(widget_ptr(new image_widget(player::victory_points_icon())));
			g->add_col(widget_ptr(new label("VP", graphics::color_white())));
		}

		const int start_devel_types = neffect;
		for(int n = 0; n != sizeof(devel_types)/sizeof(*devel_types); ++n) {
			if(effect_used != neffect++) {
				g->add_col(widget_ptr(new image_widget(player::development_icon(devel_types[n]))));
				g->add_col(widget_ptr(new label("Develop", graphics::color_white())));
			}
		}

		const int start_resource_types = neffect;
		for(int n = 0; n != 3; ++n) {
			if(effect_used != neffect++) {
				g->add_col(widget_ptr(new image_widget(player::development_icon(static_cast<DEVELOPMENT_TYPE>(n)))));
				g->add_col(widget_ptr(new label("Produce", graphics::color_white())));
			}
		}

		g->set_show_background(true);
		g->allow_selection();
		g->register_selection_callback(boost::bind(&game_dialog::make_selection, &dia, _1));
		d->add_widget(g);

		int nitem = dia.modal_selection();
		if(effect_used != -1 && nitem >= effect_used) {
			++nitem;
		}

		if(nitem == 0) {
			p.get_victory_points(1);
		} else if(nitem >= start_devel_types && nitem < start_resource_types) {
			p.develop(devel_types[nitem - start_devel_types], 1);
		} else {
			p.produce(static_cast<RESOURCE_TYPE>(nitem - start_resource_types));
		}

		effect_used = nitem;
	}
}
BUILDING_AI(workshop) { return 100; }
BUILDING_DEF(school,         1,       0,    1,   0,       3,   1, false,
  "Allows culture to grow beyond 4") {
	p.set_max_culture_level(6);
	if(p.has_building(get_building_id("library"))) {
		p.set_max_culture_level(9);
	}
}
BUILDING_AI(school) {
	const int headroom = 3 - p.development(DEVELOP_CULTURE);
	return 500 - headroom*100;
}

BUILDING_DEF(army,           1,       1,    0,   0,       3,   1, false,
  "Allows military to grow beyond 6 (must also have the barracks)") {
	if(p.has_building(get_building_id("barracks"))) {
		p.set_max_military_level(7);
	}
}
BUILDING_AI(army) { return 600; }
BUILDING_DEF(tholos,         1,       2,    0,   0,       3,   2, false,
   "Requisites to get extra priests are reduced by 1") {
	p.set_population_priest_requirement(4);
	p.set_culture_priest_requirement(6);
}
BUILDING_AI(tholos) { return 600; }
BUILDING_DEF(aqueduct,       2,       0,    1,   0,       3,   2, false,
   "When you produce, you may use the square after the one\noccupied by the progress marker.") {
	p.set_production_slider_increase(1);
}
BUILDING_AI(aqueduct) { return 600; }
BUILDING_DEF(fleet,          0,       1,    2,   0,       3,   2, false,
   "+1 in battle, you win ties.") {
	p.add_military_bonus(1);
	p.set_military_win_tie();
}
BUILDING_AI(fleet) { return 600; }
BUILDING_DEF(hellenic_alliance,
                             2,       2,    2,   0,       3,   1, false,
   "When you send a priest on a B, you get +1VP") {
	p.add_victory_points_for_beta(1);
}
BUILDING_AI(hellenic_alliance) { return 600; }
BUILDING_DEF(trade_route,    0,       0,    0,   2,       3,   0, false,
   "In the upkeep phase, you may exchange one resource\nfor the other two") {
	p.set_can_exchange_resources();
}
BUILDING_AI(trade_route) { return 600; }
BUILDING_DEF(temple_hestia,  0,       0,    0,   3,       3,   4, true,
    "(no effect)") {
	//no effect.
}
BUILDING_AI(temple_hestia) { return 600; }
BUILDING_DEF(temple_demetra, 4,       0,    0,   0,       4,   1, true,
   "+1 grain development, produce grain\nAllows use of grain in place of any other resource") {
	p.develop(DEVELOP_GRAIN, 1);
	p.produce(RESOURCE_GRAIN);
	p.set_resource_as_wildcard(RESOURCE_GRAIN);
}
BUILDING_AI(temple_demetra) { return 600; }
BUILDING_DEF(temple_artemis, 0,       4,    0,   0,       4,   1, true,
   "+1 venison development, produce venison\nAllows use of venison in place of any other resource") {
	p.develop(DEVELOP_VENISON, 1);
	p.produce(RESOURCE_VENISON);
	p.set_resource_as_wildcard(RESOURCE_VENISON);
}
BUILDING_AI(temple_artemis) { return 600; }
BUILDING_DEF(temple_poseidon,0,       0,    4,   0,       4,   1, true,
   "+1 fish development, produce fish\nAllows use of fish in place of any other resource") {
	p.develop(DEVELOP_FISH, 1);
	p.produce(RESOURCE_FISH);
	p.set_resource_as_wildcard(RESOURCE_FISH);
}
BUILDING_AI(temple_poseidon) { return 600; }
BUILDING_DEF(temple_hera,    0,       0,    0,   4,       4,   2, true,
    "The B effect of Hera is the same as her A effect for you") {
	p.set_temple(deity::get_by_name("hera")->number());
}
BUILDING_AI(temple_hera) { return 600; }
BUILDING_DEF(library,        0,       2,    0,   0,       5,   1, false,
    "Allows culture to grow beyond 7 (must have the school)") {
	if(p.has_building(get_building_id("school"))) {
		p.set_max_culture_level(9);
	}
}
BUILDING_AI(library) { return 600; }
BUILDING_DEF(goldsmith,      2,       0,    0,   0,       5,   1, false,
   "In the upkeep phase, you may pay a resource to get a VP") {
	p.set_can_exchange_resource_for_vp();
}
BUILDING_AI(goldsmith) { return 600; }
BUILDING_DEF(bouleuterion,   0,       0,    2,   0,       5,   1, false,
   "In the upkeep phase, you may pay a resource to get a development") {
	p.set_can_exchange_resource_for_development();
}
BUILDING_AI(bouleuterion) { return 600; }
BUILDING_DEF(philosophy,     0,       0,    0,   0,       5,   1, false,
   "Each opponent may get a resource of their choice.\nIf they do, you get the same resource.") {
	using namespace gui;
	for(int n = 1; n < p.get_game().players().size(); ++n) {
		player_ptr other = p.get_game().players()[(p.number()+n)%p.get_game().players().size()];
		if(other->is_local_human()) {
			game_dialog& dia = game_dialog::get();
			dialog* d = dia.set_message(&p, formatter() << player::get_name(p.number()) << " discovers philosophy.\nChoose one resource for them and you\nto receive.");
			grid_ptr g(new grid(1));

			for(int n = 0; n != 3; ++n) {
				DEVELOPMENT_TYPE type = static_cast<DEVELOPMENT_TYPE>(n);
				g->add_col(widget_ptr(new image_widget(player::development_icon(type))));
			}

			g->add_col(widget_ptr(new label("Pass", graphics::color_white())));

			g->allow_selection();
			g->register_selection_callback(boost::bind(&game_dialog::make_selection, &dia, _1));
			d->add_widget(g);

			const int nresource = dia.modal_selection();
			if(nresource < 3) {
				p.get_resource(nresource);
				other->get_resource(nresource);
			}

			dia.clear_message();
		} else {
			int nresource = rng::generate()%3;
			p.get_resource(nresource);
			other->get_resource(nresource);
		}
	}
}
BUILDING_AI(philosophy) { return 600; }
BUILDING_DEF(farm,           0,       2,    2,   0,       5,   2, false,
  "+1 grain development, +1 venison development, +1 fish development") {
	p.develop(DEVELOP_GRAIN, 1);
	p.develop(DEVELOP_VENISON, 1);
	p.develop(DEVELOP_FISH, 1);
}
BUILDING_AI(farm) { return 600; }
BUILDING_DEF(peloponnesian_league,
                             2,       0,    2,   0,       5,   2, false,
  "+2 military") {
	p.develop(DEVELOP_MILITARY, 2);
}
BUILDING_AI(peloponnesian_league) { return 600; }
BUILDING_DEF(parthenon,      1,       0,    3,   0,       5,   3, false,
  "+1 culture. Use tributes anytime.") {
	p.develop(DEVELOP_CULTURE, 1);
	p.set_can_use_tribute();
}
BUILDING_AI(parthenon) { return 600; }
BUILDING_DEF(delian_league,  2,       2,    0,   0,       5,   2, false,
  "When you win a war, if your opponent's resources are exhausted,\nyou may take additional resources from the reserve.") {
	p.set_can_take_tribute_from_reserve();
}
BUILDING_AI(delian_league) { return 600; }
BUILDING_DEF(amphitheatre,   3,       1,    0,   0,       5,   3, false,
  "In the upkeep phase, if you have at least two priests on a B you get a VP") {
	p.set_vp_for_priests_on_beta();
}
BUILDING_AI(amphitheatre) { return 600; }
BUILDING_DEF(achaean_league, 0,       3,    1,   0,       5,   3, false,
  "In the upkeep phase, if you have at least two priests on a B, you get a development") {
	p.set_development_for_priests_on_beta();
}
BUILDING_AI(achaean_league) { return 600; }
BUILDING_DEF(temple_hades,   1,       1,    1,   0,       5,   2, true,
  "Each opponent regresses two developments, or loses 3VP") {
	for(int n = 1; n != p.get_game().players().size(); ++n) {
		player_ptr other = p.get_game().players()[(p.number()+n)%p.get_game().players().size()];
		if(other->is_local_human()) {
			using namespace gui;
			game_dialog& dia = game_dialog::get();
			dialog* d = dia.set_message(&p, formatter() << player::get_name(p.number()) << " has built the Temple of Hades.\nYou must choose to lose three\nvictory points, or two developments.");
			grid_ptr g(new grid(1));
			g->add_col(widget_ptr(new image_widget(player::victory_points_icon())));
			g->add_col(widget_ptr(new label("Development", graphics::color_white())));
			g->allow_selection();
			g->register_selection_callback(boost::bind(&game_dialog::make_selection, &dia, _1));
			d->add_widget(g);
			const int nitem = dia.modal_selection();

			if(nitem == 0) {
				other->get_victory_points(-3);
			} else {
				for(int n = 0; n != 2; ++n) {
					std::vector<DEVELOPMENT_TYPE> choices;
					dialog* d = dia.set_message(&p, "Choose the development to regress");
					grid_ptr g(new grid(1));
					for(int n = 0; n != 6; ++n) {
						const DEVELOPMENT_TYPE type = static_cast<DEVELOPMENT_TYPE>(n);
						if(other->development(type) == 0) {
							continue;
						}

						g->add_col(widget_ptr(new image_widget(player::development_icon(type))));
						choices.push_back(type);
					}

					if(choices.empty()) {
						break;
					}
					g->allow_selection();
					g->register_selection_callback(boost::bind(&game_dialog::make_selection, &dia, _1));
					d->add_widget(g);
	
					const int nitem = dia.modal_selection();
					other->develop(choices[nitem], -1);
				}
			}

			dia.clear_message();
		} else {
			other->get_victory_points(-3);
		}
	}
}
BUILDING_AI(temple_hades) { return 600; }
BUILDING_DEF(temple_apollo,  0,       0,    0,   5,       5,   3, true,
   "Immunity to the plague") {
	p.set_temple(deity::get_by_name("apollo")->number());
}
BUILDING_AI(temple_apollo) { return 600; }
BUILDING_DEF(temple_zeus,    1,       1,    1,   0,       6,   1, true,
   "The B effect of Zeus is the same as his A effect for you") {
	p.set_temple(deity::get_by_name("zeus")->number());
}
BUILDING_AI(temple_zeus) { return 600; }
//BUILDING_DEF(democracy,      0,       0,    0,   0,       6,   1, false) {
//}
//BUILDING_DEF(acropolis,      1,       1,    1,   1,       7,   2, false) {
//}
BUILDING_DEF(trispastos,     0,       0,    0,   3,       7,   0, false,
  "You may build for free two of the first seven buildings") {
	std::vector<int> options;
	for(int n = 0; n != 7; ++n) {
		if(!p.has_building(n)) {
			options.push_back(n);
		}
	}

	if(options.size() <= 2 || p.is_local_ai()) {
		for(int n = 0; n < options.size() && n < 2; ++n) {
			p.build_building(options[n]);
		}

		return;
	}

	using namespace gui;
	game_dialog& dia = game_dialog::get();

	for(int n = 0; n != 2; ++n) {
		dialog* d = dia.set_message(&p, "Choose what to build for free");
		grid_ptr g(new grid(1));

		g->add_col(widget_ptr(new label(building::get_building(options[n])->name(), graphics::color_white())));

		g->allow_selection();
		g->register_selection_callback(boost::bind(&game_dialog::make_selection, &dia, _1));
		d->add_widget(g);

		const int nitem = dia.modal_selection();
		p.build_building(options[nitem]);
		options.erase(options.begin() + nitem);
	}

	dia.clear_message();
}
BUILDING_AI(trispastos) { return 600; }
BUILDING_DEF(marble_quarry,  3,       0,    1,   0,       8,   2, false,
  "Each building costs one less resource") {
	p.add_free_resource_toward_building();
}
BUILDING_AI(marble_quarry) { return 600; }
//BUILDING_DEF(oracle,         1,       3,    0,   0,       8,   3, false) {
//}
BUILDING_DEF(colony,         0,       1,    3,   0,       8,   3, false,
  "Gives three developments (on different tracks)") {
	std::vector<DEVELOPMENT_TYPE> types_used;
	for(int n = 0; n != 3; ++n) {
		std::vector<DEVELOPMENT_TYPE> options;
		for(int m = 0; m != 6; ++m) {
			DEVELOPMENT_TYPE t = static_cast<DEVELOPMENT_TYPE>(m);
			const int current = p.development(t);
			if(current > p.max_development(t) || current > p.max_building_development(t) || std::count(types_used.begin(), types_used.end(), t)) {
				continue;
			}

			options.push_back(t);
		}

		if(options.empty()) {
			break;
		}

		if(p.is_local_ai()) {
			const int result = rng::generate()%options.size();
			const DEVELOPMENT_TYPE type = options[result];
			types_used.push_back(type);
			p.develop(type, 1);
		} else {
			using namespace gui;
			game_dialog& dia = game_dialog::get();
			dialog* d = dia.set_message(&p, "Choose which area to develop\nwith your colony");
			grid_ptr g(new grid(1));
			foreach(DEVELOPMENT_TYPE type, options) {
				g->add_col(widget_ptr(new image_widget(player::development_icon(type))));
			}
			g->allow_selection();
			g->register_selection_callback(boost::bind(&game_dialog::make_selection, &dia, _1));
			d->add_widget(g);

			const int nitem = dia.modal_selection();
			p.develop(options[nitem], 1);
			types_used.push_back(options[nitem]);

			dia.clear_message();
		}
	}
}
BUILDING_AI(colony) { return 600; }
BUILDING_DEF(wonder,         1,       1,    1,   1,       8,   4, false,
  "When built, your supply of priests is replenished") {
	for(int n = 0; n != deity::num_deities(); ++n) {
		const_deity_ptr d = deity::get(n);
		std::vector<int>::const_iterator i = std::find(d->priests().begin(), d->priests().end(), p.number());
		if(i != d->priests().begin() && i != d->priests().end()) {
			p.get_priest_back();
		}
	}
}
BUILDING_AI(wonder) { return 600; }

}

std::string resource_name(int nresource)
{
	switch(nresource) {
	case RESOURCE_GRAIN: return "Grain";
	case RESOURCE_VENISON: return "Venison";
	case RESOURCE_FISH: return "Fish";
	default: return "None";
	}
}

int building::num_buildings()
{
	return all_buildings.size();
}

const_building_ptr building::get_building(int id)
{
	return all_buildings[id];
}

int building::get_building_id(const std::string& name)
{
	return building_ids[name];
}

building::building(const std::string& name, int grain, int venison, int fish, int wild, int culture, int points, bool unique, const std::string& help)
  : id_(-1), name_(name), unique_(unique), culture_(culture), points_(points), ninstances_(0), help_(help)
{
	cost_[RESOURCE_GRAIN] = grain;
	cost_[RESOURCE_VENISON] = venison;
	cost_[RESOURCE_FISH] = fish;
	cost_[NUM_RESOURCES] = wild;
}

building::~building()
{
}

void building::build(player& p) const
{
	++ninstances_;
	on_build(p);
}
