#include "asserts.hpp"
#include "deity.hpp"
#include "foreach.hpp"
#include "game.hpp"
#include "player.hpp"

namespace {

std::vector<deity_ptr> pantheon;
int add_deity(deity_ptr d)
{
	d->set_number(pantheon.size());
	pantheon.push_back(d);
	return pantheon.size();
}

#define DEFINE_DEITY(name, help) \
class name##_deity : public deity { \
public: \
	name##_deity() : deity(#name, help) {} \
	void apply_effects(player& p, int power); \
	int ai_rating(const player& p, int power) const; \
}; \
const int name##_deity_variable = add_deity(deity_ptr(new name##_deity)); \
void name##_deity::apply_effects(player& p, int power)

#define DEITY_AI(name) int name##_deity::ai_rating(const player& p, int power) const

DEFINE_DEITY(demeter, "A: produce grain and develop grain\nB: produce grain or develop grain") {
	if(power == 2) {
		p.develop(DEVELOP_GRAIN, 1);
		p.produce(RESOURCE_GRAIN);
	} else {
		p.develop_or_produce(RESOURCE_GRAIN);
	}
}

DEITY_AI(demeter) {
	const int develop_rating = p.ai_rate_development(DEVELOP_GRAIN);
	const int produce_rating = p.ai_rate_production(RESOURCE_GRAIN);
	if(power == 2) {
		return develop_rating + produce_rating;
	} else {
		return develop_rating > produce_rating ? develop_rating : produce_rating;
	}
}

DEFINE_DEITY(artemis, "A: produce venison and develop venison\nB: produce venison or develop venison") {
	if(power == 2) {
		p.develop(DEVELOP_VENISON, 1);
		p.produce(RESOURCE_VENISON);
	} else {
		p.develop_or_produce(RESOURCE_VENISON);
	}
}

DEITY_AI(artemis) {
	const int develop_rating = p.ai_rate_development(DEVELOP_VENISON);
	const int produce_rating = p.ai_rate_production(RESOURCE_VENISON);
	if(power == 2) {
		return develop_rating + produce_rating;
	} else {
		return develop_rating > produce_rating ? develop_rating : produce_rating;
	}
}

DEFINE_DEITY(poseidon, "A: produce fish and develop fish\nB: produce fish or develop fish") {
	if(power == 2) {
		p.develop(DEVELOP_FISH, 1);
		p.produce(RESOURCE_FISH);
	} else {
		p.develop_or_produce(RESOURCE_FISH);
	}
}

DEITY_AI(poseidon) {
	const int develop_rating = p.ai_rate_development(DEVELOP_FISH);
	const int produce_rating = p.ai_rate_production(RESOURCE_FISH);
	if(power == 2) {
		return develop_rating + produce_rating;
	} else {
		return develop_rating > produce_rating ? develop_rating : produce_rating;
	}
}

DEFINE_DEITY(athena, "A: +2 culture\nB: +1 culture") {
	if(p.choosing_deity() && p.has_temple(number())) {
		p.add_free_development_choice(1);
		if(p.development(DEVELOP_CULTURE) >= 5) {
			foreach(player_ptr enemy, p.get_game().players()) {
				if(enemy.get() != &p) {
					enemy->add_free_regression_choice(1);
				}
			}
		}
	}
	p.develop(DEVELOP_CULTURE, power);
}

DEITY_AI(athena) {
	return p.ai_rate_development(DEVELOP_CULTURE, power);
}
DEFINE_DEITY(aphrodite, "A: +2 population\nB: +1 population") {
	p.develop(DEVELOP_POPULATION, power);
	if(p.choosing_deity() && p.has_temple(number())) {
		p.add_free_resource_choice(p.development(DEVELOP_POPULATION) >= 9 ? 3 : 2);
	}
}
DEITY_AI(aphrodite) {
	return p.ai_rate_development(DEVELOP_POPULATION, power);
}
DEFINE_DEITY(ares, "A: choose twice: declare war | +1 military\nB: choose once: declare war | +1 military") {
	p.choose_wars(power);
}

DEITY_AI(ares) {
	return p.ai_rate_development(DEVELOP_MILITARY, power);
}

class apollo_deity : public deity {
public:
	apollo_deity() : deity("apollo", "A: +2 VP or +1 VP and unleash the plague\nB: +1 VP or immunity from the plague\nThe player on A chooses if Apollo will give victory points\nor will bring down the plague.\nIf the plague is chosen, all players\nwho do not send a priest to Apollo will\nlose one third of their population (rounded down)"), last_plague_(0), plague_unleashed_(false) {}
	void apply_effects(player& p, int power);
	int unleash_plague();
	int plague_unleashed() const;
	void clear_plague() { plague_unleashed_ = false; }
	int ai_rating(const player& p, int power) const;
private:
	int last_plague_;
	bool plague_unleashed_;
};
const int apollo_deity_variable = add_deity(deity_ptr(new apollo_deity));

int apollo_deity::ai_rating(const player& p, int power) const
{
	if(power == 2) {
		const int rate_plague = p.ai_rate_unleash_plague() + p.ai_rate_victory_points(1);
		const int rate_points = p.ai_rate_victory_points(2);
		return rate_plague > rate_points ? rate_plague : rate_points;
	} else if(plague_unleashed()) {
		return p.ai_rate_plague_protection();
	} else {
		return p.ai_rate_victory_points(1);
	}
}

void apollo_deity::apply_effects(player& p, int power) {
	if(power == 2) {
		p.choose_plague();
	} else {
		if(!plague_unleashed()) {
			p.get_victory_points(1);
		}
	}
}

int apollo_deity::unleash_plague() {
	plague_unleashed_ = true;
	++last_plague_;
	return last_plague_;
}

int apollo_deity::plague_unleashed() const {
	return plague_unleashed_ ? last_plague_ : 0;
}

DEFINE_DEITY(hephaestus, "A: Build two buildings\nB: Build one building") {
	p.choose_building(power);
}

DEITY_AI(hephaestus) {
	int best_rating = 0;
	for(int n = 0; n != building::num_buildings(); ++n) {
		if(p.can_afford_building(n)) {
			const int rating = building::get_building(n)->ai_evaluate(p);
			if(rating > best_rating) {
				best_rating = rating;
			}
		}
	}

	return best_rating;
}

DEFINE_DEITY(zeus, "A: Choose two other Deities and get their B effect\nB: Choose one other Deity and get their B effect\nHera may not be chosen") {
	if(power == 1 && p.has_temple(number())) {
		power = 2;
	}
	p.choose_other_deities(power);
}

DEITY_AI(zeus) {
	std::vector<int> ratings;
	for(int n = 0; n < number(); ++n) {
		ratings.push_back(deity::get(n)->ai_rating(p, 1));
	}

	std::sort(ratings.begin(), ratings.end());
	int result = 0;
	for(int n = 0; n != power; ++n) {
		result += ratings.back();
		ratings.pop_back();
	}

	return result;
}

DEFINE_DEITY(hera, "A: Get two developments of your choice, or produce two resources of your choice\nB: Get a development of your choice, or produce a resource of your choice") {
	if(power == 1 && p.has_temple(number())) {
		power = 2;
	}

	p.develop_or_produce(NUM_RESOURCES, power);
}

DEITY_AI(hera) {
	std::vector<int> rate_devel, rate_produce;
	for(int n = 0; n != 3; ++n) {
		rate_produce.push_back(p.ai_rate_production(n));
	}

	for(int n = 0; n != 6; ++n) {
		rate_devel.push_back(p.ai_rate_development(n));
	}

	int sum_devel = 0, sum_produce = 0;
	for(int n = 0; n != power; ++n) {
		sum_devel += rate_devel.back();
		sum_produce += rate_produce.back();
		rate_devel.pop_back();
		rate_produce.pop_back();
	}

	return sum_devel > sum_produce ? sum_devel : sum_produce;
}

}

int deity::num_deities()
{
	return pantheon.size();
}

deity_ptr deity::get(int id)
{
	return pantheon[id];
}

deity_ptr deity::get_by_name(const std::string& name)
{
	for(int n = 0; n != num_deities(); ++n) {
		if(name == pantheon[n]->name()) {
			return pantheon[n];
		}
	}

	ASSERT_LOG(false, "COULD NOT FIND DEITY: '" << name << "'");

	return deity_ptr();
}

deity::deity(const char* name, const char* help) : name_(name), help_(help)
{
}

deity::~deity()
{
}

int deity::unleash_plague()
{
	ASSERT_LOG(false, "Unleash plague on deity that doesn't have it");
	return 0;
}

namespace {
std::vector<int> active_pantheon;
}

void deity::set_active(int ndeity)
{
	clear_active();
	add_active(ndeity);
}

void deity::add_active(int ndeity)
{
	active_pantheon.push_back(ndeity);
}

void deity::clear_active()
{
	active_pantheon.clear();
}

const std::vector<int>& deity::active()
{
	return active_pantheon;
}

void deity::add_priest(int n)
{
	std::cerr << "ADD PRIEST " << n << " TO " << number() << " IN SLOT " << priests_.size() << "\n";
	ASSERT_LOG(std::count(priests_.begin(), priests_.end(), n) == 0, "TRIED TO ADD PRIEST " << n << " MULTIPLE TIMES!");
	priests_.push_back(n);
}
