
/*
   Copyright (C) 2007 by David White <dave.net>
   Part of the Silver Tree Project

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 or later.
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.

   See the COPYING file for more details.
*/
#include "SDL.h"

#include <iostream>
#include <limits.h>

#include "dialog.hpp"
#include "font.hpp"
#include "foreach.hpp"
#include "frame_rate_handler.hpp"
#include "raster.hpp"
#include "surface_cache.hpp"
#include "texture.hpp"
#include "tooltip.hpp"

namespace gui {

dialog::dialog(int x, int y, int w, int h)
  : opened_(0), cancelled_(false), clear_bg_(true), padding_(10),
    add_x_(0), add_y_(0), bg_alpha_(1.0), pump_(NULL)
{
	set_loc(x,y);
	set_dim(w,h);

	if(x == 0 && y == 0 && w == graphics::screen_width() && h == graphics::screen_height()) {
		bg_ = graphics::texture::get_frame_buffer();
	}
}

void dialog::size_to_fit()
{
	if(widgets_.empty()) {
		return;
	}

	int top = INT_MIN, bottom = INT_MIN, left = INT_MIN, right = INT_MIN;

	foreach(const widget_ptr& w, widgets_) {
		if(left == INT_MIN || w->x() < left) {
			left = w->x();
		}

		if(right == INT_MIN || w->x() + w->width() > right) {
			right = w->x() + w->width();
		}

		if(top == INT_MIN || w->y() < top) {
			top = w->y();
		}

		if(bottom == INT_MIN || w->y() + w->height() > bottom) {
			bottom = w->y() + w->height();
		}
	}

	//we set the dimensions such that the border on the right/bottom is the
	//same as at the top/left.
	set_dim(right + left, bottom + top);
}

dialog& dialog::add_widget(widget_ptr w, dialog::MOVE_DIRECTION dir)
{
	add_widget(w, add_x_, add_y_, dir);
	return *this;
}

dialog& dialog::add_widget(widget_ptr w, int x, int y,
                           dialog::MOVE_DIRECTION dir)
{
	w->set_loc(x,y);
	widgets_.push_back(w);
    register_listener(w);
	switch(dir) {
	case MOVE_DOWN:
		add_x_ = x;
		add_y_ = y + w->height() + padding_;
		break;
	case MOVE_RIGHT:
		add_x_ = x + w->width() + padding_;
		add_y_ = y;
		break;
	}
	return *this;
}

void dialog::remove_widget(widget_ptr w)
{
    deregister_listener(w);
	widgets_.erase(std::remove(widgets_.begin(),widgets_.end(),w),
	               widgets_.end());
}

void dialog::clear() { 
    foreach(widget_ptr w, widgets_) {
        deregister_listener(w);
    }
    widgets_.clear(); 
	add_x_ = add_y_ = 0;
}

void dialog::replace_widget(widget_ptr w_old, widget_ptr w_new)
{
	int x = w_old->x();
	int y = w_old->y();
	int w = w_old->width();
	int h = w_old->height();

	std::replace(widgets_.begin(), widgets_.end(), w_old, w_new);

	w_new->set_loc(x,y);
	w_new->set_dim(w,h);

    register_listener(w_new);
    deregister_listener(w_old);
}

void dialog::show() {
	opened_ = true;
	set_visible(true);
}

void dialog::show_modal_pump()
{
	input::pump pump;
    pump.register_listener(this);
	pump_ = &pump;
	show_modal();
	pump_ = NULL;
}

void dialog::show_modal()
{
	if(!pump_) {
		show_modal_pump();
		return;
	}

	frame_rate_handler fps;
	const int opened_level = opened_;
	opened_++;
	cancelled_ = false;

	while(opened_ > opened_level && pump_->process()) {
		do_processing();

		prepare_draw();
		draw();
		gui::draw_tooltip();
		complete_draw();

		fps.next_frame();
	}
}

void dialog::prepare_draw()
{
	graphics::prepare_raster();
	if(clear_bg()) {
		glClearColor(0.0,0.0,0.0,0.0);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	}
}

void dialog::complete_draw()
{
	SDL_GL_SwapBuffers();
}

void dialog::handle_draw_children() const {
	glPushMatrix();
	glTranslatef(x(),y(),0.0);
	foreach(const widget_ptr& w, widgets_) {
		w->draw();
	}
	glPopMatrix();
}

void dialog::handle_draw() const
{
	if(clear_bg()) {
		using namespace graphics;
		texture panel_background(texture::get("gui/opaque-background.png"));
		texture panel_top(texture::get("gui/opaque-border-top.png"));
		texture panel_bottom(texture::get("gui/opaque-border-bottom.png"));
		texture panel_left(texture::get("gui/opaque-border-left.png"));
		texture panel_right(texture::get("gui/opaque-border-right.png"));
		texture panel_top_left(texture::get("gui/opaque-border-topleft.png"));
		texture panel_top_right(texture::get("gui/opaque-border-topright.png"));
		texture panel_bottom_left(texture::get("gui/opaque-border-botleft.png"));
		texture panel_bottom_right(texture::get("gui/opaque-border-botright.png"));

		SDL_Rect rect = {x(),y(),width(),height()};
		blit_texture(panel_background, x(), y(), width(), height(),
		               0.0, 0.0, 0.0,
					   width()/GLfloat(panel_background.width()),
					   height()/GLfloat(panel_background.height()));
		blit_texture(panel_top, x(), y() - panel_top.height(),
		               width(), panel_top.height(), 0.0, 0.0, 0.0,
					   width()/GLfloat(panel_top.width()), 1.0);
		blit_texture(panel_bottom, x(), y() + height(),
		               width(), panel_bottom.height(), 0.0, 0.0, 0.0,
					   width()/GLfloat(panel_bottom.width()), 1.0);
		blit_texture(panel_left, x() - panel_left.width(), y(),
		               panel_left.width(), height(), 0.0, 0.0, 0.0,
					   1.0, height()/GLfloat(panel_left.height()));
		blit_texture(panel_right, x() + width(), y(),
		               panel_right.width(), height(), 0.0, 0.0, 0.0,
					   1.0, height()/GLfloat(panel_right.height()));

		blit_texture(panel_top_left, x() - panel_top_left.width(),
		                             y() - panel_top_left.height(),
									 panel_top_left.width(),
									 panel_top_left.height());
		blit_texture(panel_top_right, x() + width(),
		                              y() - panel_top_left.height(),
									  panel_top_right.width(),
									  panel_top_right.height());
		blit_texture(panel_bottom_left, x() - panel_bottom_left.width(),
		                                y() + height(),
									     panel_bottom_left.width(),
									     panel_bottom_left.height());
		blit_texture(panel_bottom_right, x() + width(),
		                                 y() + height(),
									     panel_bottom_right.width(),
									     panel_bottom_right.height());

		//fade effect for fullscreen dialogs
		if(false && bg_.valid()) {
			if(bg_alpha_ > 0.25) {
				bg_alpha_ -= 0.05;
			}
			glColor4f(1.0, 1.0, 1.0, bg_alpha_);
			graphics::blit_texture(bg_, x(), y(), width(), height(), 0.0, 0.0, 1.0, 1.0, 0.0);
			glColor4f(1.0, 1.0, 1.0, 1.0);
		}

	}
	handle_draw_children();
}

bool dialog::process_event(const SDL_Event& ev, bool claimed) {
    return widget::process_event(ev, claimed);
}

bool dialog::handle_event_children(const SDL_Event &event, bool claimed) {
	SDL_Event ev = event;
	normalize_event(&ev);
    return input::listener_container::process_event(ev, claimed);
}

bool dialog::handle_event(const SDL_Event& event, bool claimed)
{
    if(!claimed && opened_) {
        if(event.type == SDL_KEYDOWN &&
           (event.key.keysym.sym == SDLK_SPACE ||
            event.key.keysym.sym == SDLK_RETURN)) {
            close();
			cancelled_ = false;
            claimed = true;
        } else if(event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_ESCAPE) {
			close();
			cancelled_ = true;
			claimed = true;
		}
    }
    claimed |= handle_event_children(event, claimed);

	if(!claimed) {
		switch(event.type) {
		//if it's a mouse button up or down and it's within the dialog area,
		//then we claim it because nobody else should get it.
		case SDL_MOUSEBUTTONDOWN:
		case SDL_MOUSEBUTTONUP: {
			int mousex, mousey;
			SDL_GetMouseState(&mousex, &mousey);
			if(mousex >= x() && mousex < x() + width() &&
			   mousey >= y() && mousey < y() + height()) {
				claimed = true;
			}
			break;
		}
		}
	}
	return claimed;
}


}
