#include <iostream>

#include <boost/bind.hpp>

#include "asserts.hpp"
#include "building_widget.hpp"
#include "button.hpp"
#include "client_network.hpp"
#include "deity.hpp"
#include "deity_widget.hpp"
#include "foreach.hpp"
#include "formatter.hpp"
#include "game.hpp"
#include "game_dialog.hpp"
#include "grid_widget.hpp"
#include "image_widget.hpp"
#include "label.hpp"
#include "numeric_field_widget.hpp"
#include "player.hpp"
#include "priests_widget.hpp"
#include "random.hpp"
#include "raster.hpp"
#include "resources_widget.hpp"
#include "slider_widget.hpp"
#include "string_utils.hpp"
#include "wml_node.hpp"
#include "wml_utils.hpp"

using namespace gui;

namespace {
game_dialog* dialog_instance;
}

game_dialog& game_dialog::get()
{
	return *dialog_instance;
}

game_dialog::game_dialog(game& g)
  : dialog(0, 0, 1024, 768), game_(g), viewed_player_(g.local_player()), first_active_player_(0), active_player_(0), secondary_player_(0), priests_widget_(NULL),
   dialog_selection_(0), input_player_(0)
{
	dialog_instance = this;
	init();
	setup_active_player();
}

void game_dialog::init()
{
	using namespace gui;
	clear();
	deity_widgets_.clear();
	grid_ptr deity_grid(new gui::grid(4));
	for(int n = 0; n != deity::num_deities(); ++n) {
		deity_widget* w = new deity_widget(game_, deity::get(n), boost::bind(&game_dialog::select_deity, this, n));
		deity_widgets_.push_back(w);
		deity_grid->add_col(widget_ptr(w));
		std::cerr << "ADD DEITY " << n << "\n";
	}

	deity_grid->finish_row();
	add_widget(deity_grid);

	ASSERT_INDEX_INTO_VECTOR(viewed_player_, game_.players());
	player_ptr p = game_.players()[viewed_player_];
	int culture = 1;
	grid_ptr player_buildings(new gui::grid(6));
	for(int n = 0; n != building::num_buildings(); ++n) {
		const_building_ptr b = building::get_building(n);

		if(culture != b->culture()) {
			culture = b->culture();
			player_buildings->finish_row();
		}

		player_buildings->add_col(widget_ptr(new building_widget(p, b, boost::bind(&game_dialog::select_building, this, p, b))));
	}

	player_buildings->finish_row();

	add_widget(player_buildings, 1024 - player_buildings->width(), 0);

	grid_ptr sliders_grid(new gui::grid(1));

	for(int n = 0; n != NUM_DEVELOPMENTS; ++n) {
		sliders_grid->add_col(widget_ptr(new slider_widget(p, static_cast<DEVELOPMENT_TYPE>(n))));
	}

	add_widget(sliders_grid);

	add_widget(widget_ptr(new resources_widget(p)), 380, 500);

	priests_widget_ = new priests_widget(game_, p);
	add_widget(widget_ptr(priests_widget_), 200, 500);

	if(!players_grid_) {
		grid_ptr players_grid(new gui::grid(2));
		players_grid->allow_selection(true);
		players_grid->register_mouseover_callback(boost::bind(&game_dialog::select_viewed_player, this, _1));
		players_grid->set_show_background(true);
		foreach(player_ptr p, game_.players()) {
			players_grid->add_col(widget_ptr(new label(player::get_name(p->number()), player::get_color(p->number()), 18)));
			players_grid->add_col(widget_ptr(new numeric_field_widget(&p->victory_points())));
		}

		players_grid_ = players_grid;
	}

	add_widget(players_grid_, 10, 500);

	if(message_dialog_) {
		add_widget(message_dialog_, message_dialog_->x(), message_dialog_->y());
	}
}

void game_dialog::do_processing()
{
	if(network::connected()) {
		wml::const_node_ptr message = network::receive();
		if(message) {
			if(message->name() == "deity") {
				const int num = wml::get_int(message, "num");
				const int nplayer = wml::get_int(message, "player");
				ASSERT_GE(num, 0);
				ASSERT_LT(num, deity::num_deities());
				ASSERT_GE(nplayer, 0);
				ASSERT_LT(nplayer, game_.players().size());

				deity_selected(game_.players()[nplayer], num);
			} else if(message->name() == "building") {
				const int num = wml::get_int(message, "num");
				const int nplayer = wml::get_int(message, "player");
				ASSERT_GE(num, 0);
				ASSERT_LT(num, building::num_buildings());
				ASSERT_GE(nplayer, 0);
				ASSERT_LT(nplayer, game_.players().size());

				building_selected(game_.players()[nplayer], building::get_building(num));
			} else if(message->name() == "command") {
				const int num = wml::get_int(message, "command");
				ASSERT_GE(num, 0);
				ASSERT_LT(num, network_functions_.size());

				network_functions_[num]();
			}
		}
	}

	foreach(player_ptr p, game_.players()) {
		p->process();
	}

	foreach(scheduled_command& cmd, scheduled_commands_) {
		if(--cmd.cycles == 0) {
			cmd.cmd();
		}
	}

	while(scheduled_commands_.empty() == false && scheduled_commands_.front().cycles <= 0) {
		scheduled_commands_.pop_front();
	}

	foreach(const widget_modify_operation& op, widget_modify_queue_) {
		if(op.add) {
			add_widget(op.widget, op.x, op.y);
		} else {
			remove_widget(op.widget);
		}
	}

	widget_modify_queue_.clear();
	widget_trash_.clear();
}

void game_dialog::handle_draw() const
{
	dialog::handle_draw();

	foreach(const_player_ptr p, game_.players()) {
		foreach(const player::PriestInTransit& priest, p->priests_in_transit()) {
			const int x1 = priests_widget_->priest_x(p->number(), priest.source_pos);
			const int y1 = priests_widget_->priest_y(p->number(), priest.source_pos);
			const int x2 = deity_widgets_[priest.deity_dest]->priest_xpos(priest.deity_pos);
			const int y2 = deity_widgets_[priest.deity_dest]->priest_ypos(priest.deity_pos);

			const int x = (x2*priest.time + x1*(100-priest.time))/100;
			const int y = (y2*priest.time + y1*(100-priest.time))/100;
			graphics::draw_circle(x, y, 12, player::get_color(p->number()));
		}
	}
}

void game_dialog::select_deity(int n)
{
	using namespace gui;
	std::cerr << "SELECT DEITY: " << n << "\n";
	player_ptr p = game_.players()[(active_player_ + secondary_player_)%game_.players().size()];
	if(p->is_local_human() && !p->is_network()) {
		if(network::connected()) {
			char buf[1024];
			sprintf(buf, "[deity]\nnum=\"%d\"\nplayer=\"%d\"\n[/deity]", n, p->number());
			network::send(buf);
		}
		deity_selected(p, n);
	}
}

void game_dialog::deity_selected(player_ptr p, int n)
{
	p->select_deity(n);

	if(p->choosing_develop_or_produce()) {
		dialog* d = set_message(p.get(), "Choose whether to produce\nresources or develop.");
		d->add_widget(widget_ptr(new button(widget_ptr(new label("Develop", graphics::color_white())), network_function(boost::bind(&player::choose_develop, p)))));
		d->add_widget(widget_ptr(new button(widget_ptr(new label("Produce Resources", graphics::color_white())), network_function(boost::bind(&player::choose_produce, p)))));
	} else if(p->choosing_other_deities()) {
		if(p->choosing_other_deities() == 1) {
			set_message(p.get(), "Choose another Deity whose\nbenefit to receive.");
		} else {
			set_message(p.get(), "Choose two other Deities whose\nbenefits to receive.");
		}
	} else if(p->choosing_resource() || p->choosing_development()) {
		choose_resource(p, p->choosing_resource());
	} else if(p->choosing_plague()) {
		dialog* d = set_message(p.get(), "Choose whether Apollo should\nunleash the plague on all\nwho do not worship him.");
		d->add_widget(widget_ptr(new button(widget_ptr(new label("Plague!", graphics::color_white())), network_function(boost::bind(&player::select_plague, p, true)))));
		d->add_widget(widget_ptr(new button(widget_ptr(new label("No Plague", graphics::color_white())), network_function(boost::bind(&player::select_plague, p, false)))));
	} else if(p->choosing_building()) {
		dialog* d = set_message(p.get(), "Choose building to build or pass");
		d->add_widget(widget_ptr(new button(widget_ptr(new label("Pass", graphics::color_white())), network_function(boost::bind(&player::select_pass, p)))));
	} else if(p->choosing_number_of_wars()) {
		dialog* d = set_message(p.get(), "Choose whether to develop military\nor declare war.");
		grid_ptr g(new grid(p->choosing_number_of_wars()));

		for(int n = 0; n <= p->choosing_number_of_wars(); ++n) {
			for(int m = 0; m < p->choosing_number_of_wars(); ++m) {
				g->add_col(widget_ptr(new label(m < n ? "War!" : "Develop", graphics::color_white())));
			}
		}

		g->allow_selection();
		g->register_selection_callback(network_function_grid(boost::bind(&player::select_number_of_wars, p, _1)));
		d->add_widget(g);
	}
}

void game_dialog::select_building(player_ptr p, const_building_ptr b)
{
	using namespace gui;
	if(p->is_local_human() && !p->is_network() && p->choosing_building() && p->can_afford_building(b->id())) {
		if(network::connected()) {
			char buf[1024];
			sprintf(buf, "[building]\nnum=\"%d\"\nplayer=\"%d\"\n[/building]", b->id(), p->number());
			network::send(buf);
		}
		building_selected(p, b);
	}
}

void game_dialog::building_selected(player_ptr p, const_building_ptr b)
{
	int payment[4] = {0,0,0,0};
	std::vector<boost::array<int, 4> > solutions;
	const int nsolutions = p->can_afford_building(b->id(), &solutions);
	if(nsolutions == 1) {
		p->select_building(b->id(), solutions.front());
	} else {
		dialog* d = set_message(p.get(), "Choose which resources to use");
		grid_ptr g(new grid(solutions.front()[0] + solutions.front()[1] + solutions.front()[2]));
		for(int n = 0; n != solutions.size(); ++n) {
			for(int m = 0; m != 3; ++m) {
				for(int i = 0; i != solutions[n][m]; ++i) {
					g->add_col(widget_ptr(new image_widget(player::development_icon(static_cast<DEVELOPMENT_TYPE>(m)))));
				}
			}
		}

		g->allow_selection();
		g->register_selection_callback(network_function_grid(boost::bind(&game_dialog::select_building_payment, this, p, b, solutions, _1)));

		d->add_widget(g);
	}
}


void game_dialog::select_building_payment(player_ptr p, const_building_ptr b, const std::vector<boost::array<int, 4> >& v, int n)
{
	if(n < 0 || n >= v.size()) {
		return;
	}

	p->select_building(b->id(), v[n]);
}

void game_dialog::player_chooses_deity(int n)
{
	using namespace gui;
	player_ptr p = game_.players()[active_player_];
	if(p->choosing_other_deities()) {
		set_message(p.get(), "Choose another Deity whose\nbenefit to receive.");
		return;
	} else if(p->choosing_resource() || p->choosing_development()) {
		choose_resource(p, p->choosing_resource());
	} else if(p->choosing_building()) {
		dialog* d = set_message(p.get(), "Choose building to build or pass");
		d->add_widget(widget_ptr(new button(widget_ptr(new label("Pass", graphics::color_white())), network_function(boost::bind(&player::select_pass, p)))));
		return;
	}

	select_resource(p);
	for(int m = 0; m != game_.players().size(); ++m) {
		select_development(game_.players()[(p->number()+m)%game_.players().size()]);
	}

	clear_message();

	secondary_player_ = 1;
	schedule_command(boost::bind(&game_dialog::setup_secondary_player, this, n), 25);
}

void game_dialog::player_secondary_done(int n)
{
	deity::clear_active();

	//if the secondary player was at hera, and selected development or
	//a resource, handle that here.
	player_ptr secondary_player = game_.players()[(active_player_ + secondary_player_)%game_.players().size()];
	if(secondary_player->choosing_resource() || secondary_player->choosing_development()) {
		choose_resource(secondary_player, secondary_player->choosing_resource());
	}

	select_resource(game_.players()[(active_player_ + secondary_player_)%game_.players().size()]);
	for(int m = 0; m != game_.players().size(); ++m) {
		select_development(game_.players()[(active_player_ + secondary_player_ + m)%game_.players().size()]);
	}

	++secondary_player_;
	const int index = (active_player_ + secondary_player_)%game_.players().size();
	if(index == active_player_) {
		if(deity::get(n)->plague_unleashed()) {
			std::ostringstream s;
			s << "A plague sweeps across Greece...\n";
			foreach(player_ptr p, game_.players()) {
				s << player::get_name(p->number());
				const std::vector<int>& priests = deity::get(n)->priests();
				if(std::count(priests.begin(), priests.end(), p->number()) == 0 && !p->has_temple(n)) {
					const int before = p->development(DEVELOP_POPULATION);
					p->suffer_plague();
					const int after = p->development(DEVELOP_POPULATION);
					if(before == after) {
						s << " suffers no losses\n";
					} else {
						s << " loses " << (before - after) << " population\n";
					}
				} else {
					s << " is protected from the plague\n";
				}
			}

			modal_info_message(s.str());
			deity::get(n)->clear_plague();
		}

		next_active_player();
	} else {
		clear_message();
		schedule_command(boost::bind(&game_dialog::setup_secondary_player, this, n), 25);
	}
}

void game_dialog::declare_war(player_ptr p, int nenemy)
{
	player_ptr q = game_.players()[nenemy];
	int tribute = p->combat_strength() - q->combat_strength();
	if(tribute == 0) {
		if(!q->military_win_tie() || p->military_win_tie()) {
			tribute = 1;
		}
	}

	if(tribute < 0) {
		tribute = 0;
	}

	if(tribute > 0) {
		p->win_war();
	}

	const int resources_available = q->resources(RESOURCE_GRAIN) + q->resources(RESOURCE_VENISON) + q->resources(RESOURCE_FISH);
	if(tribute > resources_available && !p->can_take_tribute_from_reserve()) {
		tribute = resources_available;
	}

	if(tribute) {
		if(p->is_local_human()) {
			show_tribute_dialog(p, q, tribute);
			return;
		}

		modal_info_message(formatter() << player::get_name(p->number()) << " wins a war with\n" << player::get_name(q->number()) << " and gets " << tribute << " resources in tribute!");

		p->choose_tribute(q, tribute);
	}

	next_active_player();
}

void game_dialog::show_tribute_dialog(player_ptr p, player_ptr q, int tribute)
{
	const bool getting_from_reserve = (q->resources(RESOURCE_FISH) + q->resources(RESOURCE_VENISON) + q->resources(RESOURCE_GRAIN)) == 0;
	using namespace gui;
	dialog* d = set_message(p.get(), formatter() << "You have won the battle and\nmay choose " << tribute << " " << (tribute > 1 ? "resources" : "resource") << " in tribute");
	grid_ptr g(new grid(2));
	for(int n = 0; n != 3; ++n) {
		if(!getting_from_reserve && q->resources(static_cast<RESOURCE_TYPE>(n)) == 0) {
			continue;
		}
		g->add_col(widget_ptr(new button(widget_ptr(new image_widget(player::development_icon(static_cast<DEVELOPMENT_TYPE>(n)))), network_function(boost::bind(&game_dialog::get_tribute, this, p, q, n, tribute-1)))));
		g->add_col(widget_ptr(new label(formatter() << q->resources(static_cast<RESOURCE_TYPE>(n)), graphics::color_white())));
	}
	d->add_widget(g);
}

void game_dialog::select_resource(player_ptr p)
{
	if(p->number_of_resources_choosing() <= 0) {
		return;
	}

	if(p->is_local_ai()) {
		while(p->number_of_resources_choosing() > 0) {
			p->choose_resource(rng::generate()%3);
		}

		return;
	}

	while(p->number_of_resources_choosing() > 0) {
		dialog* d = set_message(p.get(), "Choose a resource");
		grid_ptr g(new grid(1));
		for(int n = 0; n != 3; ++n) {
			g->add_col(widget_ptr(new image_widget(player::development_icon(static_cast<DEVELOPMENT_TYPE>(n)))));
		}
		g->allow_selection();
		g->register_selection_callback(network_function_grid(boost::bind(&game_dialog::make_selection, this, _1)));
		d->add_widget(g);

		const int nitem = modal_selection();
		p->choose_resource(nitem);
	}

	clear_message();
}

void game_dialog::select_development(player_ptr p)
{
	if(p->number_of_developments_choosing() <= 0) {
		return;
	}

	if(p->is_local_ai()) {
		while(p->number_of_developments_choosing() > 0) {
			int best = 0;
			for(int n = 1; n != 6; ++n) {
				if(p->ai_rate_development(n) > p->ai_rate_development(best)) {
					best = n;
				}
			}

			p->choose_development(best);
		}

		while(p->number_of_regressions_choosing() > 0) {
			p->choose_regression(rng::generate()%6);
		}

		return;
	}

	while(p->number_of_developments_choosing() > 0) {
		const int choice = modal_resource_selection(p, "Choose an area to develop", 1, 6);
		p->choose_development(choice);
	}

	while(p->number_of_regressions_choosing() > 0) {
		const int choice = modal_resource_selection(p, "Choose an area to regress", 1, 6);
		p->choose_regression(choice);
	}
}

void game_dialog::exchange_resources(player_ptr p)
{
	if(p->resources(RESOURCE_FISH) + p->resources(RESOURCE_VENISON) + p->resources(RESOURCE_GRAIN) == 0) {
		return;
	}

	if(p->is_local_ai()) {
		int n = rng::generate()%3;
		while(p->resources(static_cast<RESOURCE_TYPE>(n)) == 0) {
			n = rng::generate()%3;
		}

		p->exchange_resources(n);
		return;
	}

	dialog* d = set_message(p.get(), "Do you want to trade a resource?");
	std::vector<RESOURCE_TYPE> vals;
	grid_ptr g(new grid(4));
	for(int n = 0; n != 3; ++n) {
		const RESOURCE_TYPE t = static_cast<RESOURCE_TYPE>(n);
		if(p->resources(t) == 0) {
			continue;
		}

		g->add_col(widget_ptr(new image_widget(player::development_icon(static_cast<DEVELOPMENT_TYPE>(n)))));
		g->add_col(widget_ptr(new label(" -> ", graphics::color_white())));
		for(int m = 0; m != 3; ++m) {
			if(m == n) {
				continue;
			}
			g->add_col(widget_ptr(new image_widget(player::development_icon(static_cast<DEVELOPMENT_TYPE>(m)))));
		}

		vals.push_back(t);
	}

	g->add_col(widget_ptr(new label("Pass", graphics::color_white())));
	g->finish_row();
	g->allow_selection();
	g->register_selection_callback(network_function_grid(boost::bind(&game_dialog::make_selection, this, _1)));
	d->add_widget(g);
	const int nitem = modal_selection();
	if(nitem < vals.size()) {
		p->exchange_resources(vals[nitem]);
	}

	clear_message();
}

void game_dialog::exchange_resource_for_vp(player_ptr p)
{
	if(p->resources(RESOURCE_FISH) + p->resources(RESOURCE_VENISON) + p->resources(RESOURCE_GRAIN) == 0) {
		return;
	}

	if(p->is_local_ai()) {
		return;
	}

	dialog* d = set_message(p.get(), "Do you want to trade a resource\nfor a victory point?");
	std::vector<RESOURCE_TYPE> vals;
	grid_ptr g(new grid(3));
	for(int n = 0; n != 3; ++n) {
		const RESOURCE_TYPE t = static_cast<RESOURCE_TYPE>(n);
		if(p->resources(t) == 0) {
			continue;
		}

		g->add_col(widget_ptr(new image_widget(player::development_icon(static_cast<DEVELOPMENT_TYPE>(n)))));
		g->add_col(widget_ptr(new label(" -> ", graphics::color_white())));
		g->add_col(widget_ptr(new image_widget(player::victory_points_icon())));

		vals.push_back(t);
	}

	g->add_col(widget_ptr(new label("Pass", graphics::color_white())));
	g->finish_row();
	g->allow_selection();
	g->register_selection_callback(network_function_grid(boost::bind(&game_dialog::make_selection, this, _1)));
	d->add_widget(g);
	const int nitem = modal_selection();
	if(nitem < vals.size()) {
		p->exchange_resource_for_vp(vals[nitem]);
	}

	clear_message();
}

void game_dialog::exchange_resource_for_development(player_ptr p)
{
	if(p->resources(RESOURCE_FISH) + p->resources(RESOURCE_VENISON) + p->resources(RESOURCE_GRAIN) == 0) {
		return;
	}

	if(p->is_local_ai()) {
		return;
	}

	dialog* d = set_message(p.get(), "Do you want to trade a resource\nfor development?");
	std::vector<RESOURCE_TYPE> vals;
	grid_ptr g(new grid(1));
	for(int n = 0; n != 3; ++n) {
		const RESOURCE_TYPE t = static_cast<RESOURCE_TYPE>(n);
		if(p->resources(t) == 0) {
			continue;
		}

		g->add_col(widget_ptr(new image_widget(player::development_icon(static_cast<DEVELOPMENT_TYPE>(n)))));
		vals.push_back(t);
	}

	g->add_col(widget_ptr(new label("Pass", graphics::color_white())));
	g->finish_row();
	g->allow_selection();
	g->register_selection_callback(network_function_grid(boost::bind(&game_dialog::make_selection, this, _1)));
	d->add_widget(g);
	const int nitem = modal_selection();
	if(nitem >= vals.size()) {
		return;
	}

	choose_resource(p, true);
}


void game_dialog::choose_resource(player_ptr p, bool choosing_resource, const char* message)
{

	const int num = std::max<int>(1, std::max<int>(p->choosing_resource(), p->choosing_development()));
	const int resource = modal_resource_selection(p, message ? message : (choosing_resource ? "Choose resources" : "Choose development"), num);

	p->choose_resources_to_develop_or_produce(static_cast<RESOURCE_TYPE>(resource));
}

int game_dialog::modal_resource_selection(player_ptr p, const std::string& msg, int num, int noptions)
{
	dialog* d = set_message(p.get(), msg);
	for(int n = 0; n != noptions; ++n) {
		grid_ptr g(new grid(num));

		for(int m = 0; m != num; ++m) {
			g->add_col(widget_ptr(new image_widget(player::development_icon(static_cast<DEVELOPMENT_TYPE>((n+m)%noptions)))));
		}

		widget_ptr b(new button(g, network_function(boost::bind(&game_dialog::make_selection, this, n))));
		d->add_widget(b);
	}
	const int resource = modal_selection();
	clear_message();
	return resource;
}

void game_dialog::modal_info_message(const std::string& msg)
{
	dialog* d = set_message(NULL, msg);
	d->add_widget(widget_ptr(new button(widget_ptr(new label("Continue", graphics::color_white())), boost::bind(&game_dialog::make_selection, this, 0))));
	modal_selection();
	clear_message();
}

void game_dialog::get_tribute(player_ptr p, player_ptr q, int resource, int tribute_left)
{
	p->get_tribute(q, resource);
	if(tribute_left > 0) {
		show_tribute_dialog(p, q, tribute_left);
	} else {
		next_active_player();
	}
}

void game_dialog::next_active_player()
{
	using namespace gui;
	foreach(player_ptr p, game_.players()) {
		if(p->number_of_wars()) {
			p->declare_war(network_function_grid(boost::bind(&game_dialog::declare_war, this, p, _1)));
			if(p->declaring_war()) {
				dialog* d = set_message(p.get(), "Choose who to declare war on.");
				grid_ptr g(new grid(3));

				foreach(player_ptr enemy, game_.players()) {
					if(enemy != p) {
						g->add_col(widget_ptr(new label(player::get_name(enemy->number()), graphics::color_white())));
						g->add_col(widget_ptr(new image_widget(player::development_icon(DEVELOP_MILITARY))));
						g->add_col(widget_ptr(new label(formatter() << enemy->combat_strength()+1, graphics::color_white())));
					}
				}

				g->allow_selection();
				g->register_selection_callback(network_function_grid(boost::bind(&player::make_war_declaration, p, _1)));

				d->add_widget(g);
			}

			return;
		}
	}

	int unclaimed = 0;
	for(int n = 0; n != deity::num_deities(); ++n) {
		if(deity::get(n)->priests().empty()) {
			++unclaimed;
		}
	}

	secondary_player_ = 0;

	int npass = unclaimed == 0 ? game_.players().size()+1 : 0;
	active_player_ = (active_player_+1)%game_.players().size();
	while(npass <= game_.players().size() && game_.players()[active_player_]->num_priests() == 0) {
		++npass;
		active_player_ = (active_player_+1)%game_.players().size();
	}

	if(npass <= game_.players().size()) {
		clear_message();
		schedule_command(boost::bind(&game_dialog::setup_active_player, this), 25);
	} else {
		maintenance_phase();
	}
}

void game_dialog::setup_active_player()
{
	player_ptr p = game_.players()[active_player_];

	std::cerr << "ACTIVE PLAYER: " << p->number() << "\n";

	const bool res = p->active_player_turn(boost::bind(&game_dialog::player_chooses_deity, this, _1));

	if(!p->is_local_human() || !res) {
		return;
	}

	set_message(p.get(), "Choose which temple to send a\npriest to.");
}

void game_dialog::setup_secondary_player(int ndeity)
{
	using namespace gui;
	player_ptr p = game_.players()[(active_player_ + secondary_player_)%game_.players().size()];
	std::cerr << "SECONDARY PLAYER: " << p->number() << "\n";
	const bool res = p->secondary_player_turn(ndeity, boost::bind(&game_dialog::player_secondary_done, this, _1));

	if(!p->is_local_human() || !res) {
		return;
	}

	std::cerr << "SETUP SECONDARY\n";

	deity::set_active(ndeity);

	dialog* d = set_message(p.get(), formatter() << "Send a priest to " << deity::get(ndeity)->name() << " or pass");
	d->add_widget(widget_ptr(new button(widget_ptr(new label("Pass", graphics::color_white())), network_function(boost::bind(&player::select_pass, p)))));
}

gui::dialog* game_dialog::set_message(player* p, const std::string& msg)
{
	using namespace gui;
	dialog* d = new dialog(800, 400, 260, 140);
	dialog* result_dialog = d;

	input_player_ = p;

	network_functions_.clear();

	if(p && p->is_network()) {
		d = new dialog(800, 400, 260, 140);
		d->add_widget(widget_ptr(new label(formatter() << "Waiting for " << player::get_name(p->number()) << "...", graphics::color_white())));
		network_dialog_.reset(result_dialog);
	} else {
		std::vector<std::string> v = util::split(msg, '\n');
		foreach(const std::string& s, v) {
			d->add_widget(widget_ptr(new label(s, graphics::color_white())));
		}
	}

	if(message_dialog_) {
		remove_widget(message_dialog_);
		widget_trash_.push_back(message_dialog_);
	}

	message_dialog_.reset(d);

	add_widget(message_dialog_, 300, 340);

	return result_dialog;
}

void game_dialog::clear_message()
{
	set_message(NULL, "");
}

void game_dialog::maintenance_phase()
{
	foreach(player_ptr p, game_.players()) {
		p->apply_population_limit();
	}

	foreach(player_ptr p, game_.players()) {
		while(p->total_num_resources() > 5) {
			if(p->is_local_ai()) {
				const int resource = rng::generate()%3;
				if(p->resources(static_cast<RESOURCE_TYPE>(resource)) > 0) {
					p->get_resource(resource, -1);
				}
			} else {
				const int resource = modal_resource_selection(p, "Upkeep phase: discard all resources\nmore than five. Choose which\nresource to discard");
				if(p->resources(static_cast<RESOURCE_TYPE>(resource)) > 0) {
					p->get_resource(resource, -1);
				}
			}
		}
	}

	foreach(player_ptr p, game_.players()) {
		int nbeta = 0;
		for(int n = 0; n != deity::num_deities(); ++n) {
			deity_ptr d = deity::get(n);
			std::vector<int>::const_iterator i = std::find(d->priests().begin(), d->priests().end(), p->number());
			if(i != d->priests().begin() && i != d->priests().end()) {
				++nbeta;
			}
		}

		if(nbeta == 2) {
			if(p->get_vp_for_priests_on_beta()) {
				p->get_victory_points(1);
			}

			if(p->get_development_for_priests_on_beta()) {
				choose_resource(p, false, "Choose development from the\nAchaean League");
			}
		}
	}

	foreach(player_ptr p, game_.players()) {
		if(p->can_exchange_resources()) {
			exchange_resources(p);
		}
	}

	foreach(player_ptr p, game_.players()) {
		if(p->can_exchange_resource_for_vp()) {
			exchange_resource_for_vp(p);
		}
	}

	foreach(player_ptr p, game_.players()) {
		if(p->can_exchange_resource_for_development()) {
			exchange_resource_for_development(p);
		}
	}

	foreach(player_ptr p, game_.players()) {
		p->collect_tribute();
	}

	modal_info_message("A new phase begins");

	for(int n = 0; n != deity::num_deities(); ++n) {
		deity::get(n)->clear_priests();
	}

	foreach(player_ptr p, game_.players()) {
		p->restore_priests();
	}

	int glories = 0;
	for(int n = 0; n != 6; ++n) {
		if(player::glory_owner(n) != -1) {
			++glories;
		}
	}

	if(glories >= 4) {
		std::vector<const_player_ptr> winners;
		foreach(const_player_ptr p, game_.players()) {
			if(winners.empty() || p->victory_points() > winners.front()->victory_points()) {
				winners.clear();
				winners.push_back(p);
			} else if(p->victory_points() == winners.front()->victory_points()) {
				winners.push_back(p);
			}
		}

		if(winners.size() > 1) {
			modal_info_message("The game ends in a tie!");
		} else {
			modal_info_message(formatter() << "Game over: " << player::get_name(winners.front()->number()) << " wins!");
		}
		
		throw game_over_exception();
	}

	++first_active_player_;
	active_player_ = first_active_player_%game_.players().size();
	secondary_player_ = 0;
	setup_active_player();
}

void game_dialog::select_viewed_player(int nplayer)
{
	if(nplayer < 0 || nplayer >= game_.players().size()) {
		nplayer = game_.local_player();
	}
	viewed_player_ = nplayer;
	init();
}

void game_dialog::schedule_command(boost::function<void()> cmd, int cycles)
{
	cmd();
	return;
	cycles = 25;
	scheduled_command c;
	c.cmd = cmd;
	c.cycles = cycles;
	scheduled_commands_.push_back(c);
}

void game_dialog::make_selection(int nselect)
{
	if(nselect < 0) {
		return;
	}

	dialog_selection_ = nselect;
	close();
}

int game_dialog::modal_selection()
{
	show_modal();
	return dialog_selection_;
}

namespace {

void execute_and_send_command(boost::function<void()> f, int slot)
{
	if(network::connected()) {
		char buf[1024];
		sprintf(buf, "[command]\ncommand=\"%d\"\n[/command]", slot);
		network::send(buf);
	}

	f();
}

void execute_and_send_command_grid(boost::function<void(int)> f, int slot, int n)
{
	if(network::connected()) {
		char buf[1024];
		sprintf(buf, "[command]\ncommand=\"%d\"\n[/command]", slot+n);
		network::send(buf);
	}

	f(n);
}

}

boost::function<void()> game_dialog::network_function(boost::function<void()> f)
{
	const int slot = network_functions_.size();
	network_functions_.push_back(f);
	if(input_player_ && input_player_->is_network()) {
	} else {
		return boost::bind(execute_and_send_command, f, slot);
	}
	return f;
}

boost::function<void(int)> game_dialog::network_function_grid(boost::function<void(int)> f)
{
	const int slot = network_functions_.size();
	for(int n = 0; n != 32; ++n) {
		network_functions_.push_back(boost::bind(f, n));
	}
	if(input_player_ && input_player_->is_network()) {
	} else {
		return boost::bind(execute_and_send_command_grid, f, slot, _1);
	}
	return f;
}
