#include <algorithm>
#include <numeric>

#include "asserts.hpp"
#include "deity.hpp"
#include "foreach.hpp"
#include "game.hpp"
#include "player.hpp"
#include "random.hpp"
#include "raster.hpp"

namespace {
int g_glories[6] = { -1, -1, -1, -1, -1, -1 };
}

int player::glory_owner(int n)
{
	return g_glories[n];
}

graphics::texture player::victory_points_icon()
{
	static const graphics::texture icons(graphics::texture::get("icons.png"));
	static const graphics::texture t = icons.get_portion(64, 64, 96, 96);
	return t;
}

graphics::texture player::wildcard_resource_icon()
{
	static const graphics::texture icons(graphics::texture::get("icons.png"));
	static const graphics::texture t = icons.get_portion(32, 64, 64, 96);
	return t;
}

graphics::texture player::development_icon(DEVELOPMENT_TYPE type)
{
	static const graphics::texture icons(graphics::texture::get("icons.png"));
	static const graphics::texture devel_icons[] = {
	   icons.get_portion(64, 0, 96, 32),
	   icons.get_portion(64, 32, 96, 64),
	   icons.get_portion(0, 0, 32, 32),
	   icons.get_portion(32, 32, 64, 64),
	   icons.get_portion(0, 32, 32, 64),
	   icons.get_portion(32, 0, 64, 32),
	};

	return devel_icons[type];
}

const char* player::get_name(int nplayer)
{
	static const char* names[] = {"Athens", "Sparta", "Corinth", "Argos"};
	return names[nplayer];
}

const SDL_Color& player::get_color(int nplayer)
{
	static const SDL_Color colors[] = {graphics::color_red(), graphics::color_blue(), graphics::color_green(), graphics::color_yellow()};
	return colors[nplayer];
}

int player::resource_productivity(int devel)
{
	switch(devel) {
	case 0:
	case 1:
		return 1;
	case 2:
	case 3:
		return 2;
	case 4:
	case 5:
	case 6:
		return 3;
	default:
		return 4;
	}
}

player::player(game& g, PLAYER_TYPE type, int num)
  : game_(g), type_(type), victory_points_(10), max_victory_points_payment_for_building_(0),
    number_(num), npriests_(3),
    state_(STATE_NONE), ndeities_choose_(0), nbuildings_to_build_(0), nwars_(0),
    max_culture_level_(3), max_population_level_(4), max_military_level_(2),
	building_culture_reduction_(0), production_slider_increase_(0),
	population_priest_requirement_(5), culture_priest_requirement_(7),
	military_bonus_(0), military_win_tie_(false),
	victory_points_for_beta_(0), nchoose_resources_(0),
	nchoose_developments_(0), nchoose_regressions_(0),
	gets_priest_when_win_war_(false),
	can_exchange_resources_(false),
	can_exchange_resource_for_vp_(false),
	can_exchange_resource_for_development_(false),
	free_resources_toward_building_(0),
	can_use_tribute_(false),
	can_take_tribute_from_reserve_(false),
	get_vp_for_priests_on_beta_(false),
	get_development_for_priests_on_beta_(false)
{
	for(int n = 0; n != NUM_DEVELOPMENTS; ++n) {
		development_[n] = 0;
	}
	development_[DEVELOP_POPULATION] = 1;

	for(int n = 0; n != NUM_DEVELOPMENTS; ++n) {
		development_draw_[n] = development_[n]*100;
	}

	for(int n = 0; n != NUM_RESOURCES; ++n) {
		resources_[n] = 1;
		tribute_[n] = 0;
		can_use_resource_as_wildcard_[n] = false;
	}
}

int player::max_building_development(DEVELOPMENT_TYPE type) const
{
	switch(type) {
	case DEVELOP_GRAIN:
	case DEVELOP_VENISON:
	case DEVELOP_FISH:
		return 7;
	case DEVELOP_CULTURE:
		return max_culture_level_;
	case DEVELOP_POPULATION:
		return max_population_level_;
	case DEVELOP_MILITARY:
		return max_military_level_;
	}

	ASSERT_LOG(false, "bad devel type");
	return 0;
}

int player::max_development(DEVELOPMENT_TYPE type) const
{
	switch(type) {
	case DEVELOP_GRAIN:
	case DEVELOP_VENISON:
	case DEVELOP_FISH:
		return 7;
	case DEVELOP_CULTURE:
	case DEVELOP_POPULATION:
		return 9;
	case DEVELOP_MILITARY:
		return 7;
	}

	ASSERT_LOG(false, "bad devel type");
	return 0;
}

void player::develop(DEVELOPMENT_TYPE type, int amount)
{
	development_[type] += amount;
	if(development_[type] < 0) {
		development_[type] = 0;
	}

	if(development_[type] > max_building_development(type)) {
		development_[type] = max_building_development(type);
	}

	if(development_[type] == max_development(type) && g_glories[type] == -1) {
		g_glories[type] = number();
		victory_points_ += 2;
	}
}

void player::produce(RESOURCE_TYPE type)
{
	resources_[type] += resource_productivity(development_[type]+production_slider_increase_);
}

void player::develop_or_produce(RESOURCE_TYPE type, int num)
{
	if(is_local_ai()) {
		if(type == NUM_RESOURCES) {
			int best = 0;
			for(int n = 0; n != 3; ++n) {
				const int rate_develop = ai_rate_development(n);
				const int rate_production = ai_rate_production(n);
				if(rate_develop > best) {
					best = rate_develop;
					type = static_cast<RESOURCE_TYPE>(n);
				}

				if(rate_production > best) {
					best = rate_production;
					type = static_cast<RESOURCE_TYPE>(n);
				}
			}
		}

		const int rate_develop = ai_rate_development(type);
		const int rate_production = ai_rate_production(type);
		const bool choose_develop = rate_develop > rate_production;

		int type_choice = type;
		for(int n = 0; n < num; ++n) {
			if(choose_develop) {
				develop(static_cast<DEVELOPMENT_TYPE>(type_choice), 1);
			} else {
				produce(static_cast<RESOURCE_TYPE>(type_choice));
			}

			type_choice = (type_choice+1)%3;
		}
	} else if(is_local_human()) {
		state_ = STATE_DEVELOP_OR_PRODUCE;
		resource_type_choosing_ = type;
		develop_or_produce_num_choices_ = num;
	}
}

void player::choose_develop()
{
	ASSERT_EQ(state_, STATE_DEVELOP_OR_PRODUCE);

	if(resource_type_choosing_ == NUM_RESOURCES) {
		state_ = STATE_DEVELOP_CHOICE;
		finish_turn_(current_deity_);
		return;
	}

	develop(static_cast<DEVELOPMENT_TYPE>(resource_type_choosing_), 1);

	if(ndeities_choose_) {
		state_ = STATE_CHOOSE_OTHER_DEITIES;
	} else {
		state_ = STATE_NONE;
	}

	finish_turn_(current_deity_);
}

void player::choose_produce()
{
	ASSERT_EQ(state_, STATE_DEVELOP_OR_PRODUCE);

	if(resource_type_choosing_ == NUM_RESOURCES) {
		state_ = STATE_PRODUCE_CHOICE;
		finish_turn_(current_deity_);
		return;
	}

	produce(resource_type_choosing_);
	if(ndeities_choose_) {
		state_ = STATE_CHOOSE_OTHER_DEITIES;
	} else {
		state_ = STATE_NONE;
	}

	finish_turn_(current_deity_);
}

void player::choose_resources_to_develop_or_produce(RESOURCE_TYPE resource)
{
	for(int n = 0; n != develop_or_produce_num_choices_; ++n) {
		const RESOURCE_TYPE r = static_cast<RESOURCE_TYPE>((resource+n)%3);
		if(choosing_resource()) {
			produce(r);
		} else {
			develop(static_cast<DEVELOPMENT_TYPE>(r), 1);
		}
	}

	develop_or_produce_num_choices_ = 0;
	state_ = STATE_NONE;
}

int player::total_num_resources() const
{
	int result = 0;
	for(int n = 0; n != NUM_RESOURCES; ++n) {
		result += resources_[n];
	}

	return result;
}

void player::restore_priests()
{
	npriests_ = max_priests();
}

int player::max_priests() const
{
	return 3 + (development_[DEVELOP_POPULATION] >= population_priest_requirement_ ? 1 : 0) + (development_[DEVELOP_CULTURE] >= culture_priest_requirement_ ? 1 : 0);
}

void player::get_deity_benefit(int ndeity, int power)
{
	const bool choosing_others = state_ == STATE_CHOOSE_OTHER_DEITIES;
	deity_ptr d = deity::get(ndeity);
	d->apply_effects(*this, power);

	if(!choosing_others && state_ != STATE_DEVELOP_OR_PRODUCE && state_ != STATE_CHOOSE_OTHER_DEITIES && state_ != STATE_CHOOSE_PLAGUE && state_ != STATE_CHOOSE_BUILDING && state_ != STATE_CHOOSE_WAR) {
		state_ = STATE_NONE;
		finish_turn_(current_deity_);
	}
}

void player::select_deity(int n)
{
	deity_ptr d = deity::get(n);
	if(state_ == STATE_CHOOSE_DEITY && d->priests().empty()) {
		current_deity_ = n;
		move_priest_to_deity(n);
		get_deity_benefit(n, 2);
	} else if(state_ == STATE_CHOOSE_SECONDARY && n == current_deity_) {
		move_priest_to_deity(n);
		victory_points_ += victory_points_for_beta_;
		get_deity_benefit(n, 1);
	} else if(state_ == STATE_CHOOSE_OTHER_DEITIES && std::count(deities_unavailable_.begin(), deities_unavailable_.end(), n) == 0) {
		deities_unavailable_.push_back(n);
		--ndeities_choose_;
		if(ndeities_choose_ == 0) {
			state_ = STATE_NONE;
		}
		get_deity_benefit(n, 1);
	}
}

void player::select_pass()
{
	if(is_local_human() && state_ == STATE_CHOOSE_SECONDARY) {
		state_ = STATE_NONE;
		finish_turn_(current_deity_);
	} else if(is_local_human() && state_ == STATE_CHOOSE_BUILDING) {
		state_ = STATE_NONE;
		nbuildings_to_build_ = 0;
		finish_turn_(current_deity_);
	}
}

void player::choose_other_deities(int n)
{
	state_ = STATE_CHOOSE_OTHER_DEITIES;
	if(is_local_human()) {
		deities_unavailable_.clear();
		deities_unavailable_.push_back(8);
		deities_unavailable_.push_back(9);
		ndeities_choose_ = n;
	} else if(is_local_ai()) {
		typedef std::pair<int, int> RatingOfDeity;
		std::vector<RatingOfDeity> ratings;
		for(int m = 0; m != 6; ++m) {
			const int rating = deity::get(m)->ai_rating(*this, 1);
			ratings.push_back(RatingOfDeity(rating, m));
		}

		std::sort(ratings.begin(), ratings.end());
		ratings.erase(ratings.begin(), ratings.end() - n);

		foreach(const RatingOfDeity& r, ratings) {
			state_ = STATE_CHOOSE_OTHER_DEITIES;
			get_deity_benefit(r.second, 1);
		}

		state_ = STATE_NONE;
	}
}

int player::choosing_other_deities() const
{
	return state_ == STATE_CHOOSE_OTHER_DEITIES ? ndeities_choose_ : 0;
}

void player::choose_plague()
{
	if(is_local_ai()) {
		if(ai_rate_victory_points(1) + ai_rate_unleash_plague() > ai_rate_victory_points(2)) {
			victory_points_ += 1;
			deity::get(current_deity_)->unleash_plague();
		} else {
			victory_points_ += 2;
		}

		state_ = STATE_NONE;
		return;
	}

	state_ = STATE_CHOOSE_PLAGUE;
}

void player::select_plague(bool plague)
{
	if(plague) {
		victory_points_ += 1;
		deity::get(current_deity_)->unleash_plague();
	} else {
		victory_points_ += 2;
	}

	state_ = STATE_NONE;
	finish_turn_(current_deity_);
}

void player::choose_building(int nbuildings)
{
	if(is_local_ai()) {
		while(nbuildings--) {
			int best_rating = 0, best_building = -1;
			for(int n = 0; n != building::num_buildings(); ++n) {
				if(!can_afford_building(n)) {
					continue;
				}

				const int rating = building::get_building(n)->ai_evaluate(*this);
				if(rating > best_rating) {
					best_rating = rating;
					best_building = n;
				}
			}

			if(best_building == -1) {
				break;
			}

			std::vector<boost::array<int, 4> > payment;
			can_afford_building(best_building, &payment);
			ASSERT_NE(payment.empty(), true);
			for(int n = 0; n != 3; ++n) {
				resources_[n] -= payment.front()[n];
			}

			victory_points_ -= payment.front()[3];

			build_building(best_building);
		}
		state_ = STATE_NONE;
	} else if(is_local_human()) {
		state_ = STATE_CHOOSE_BUILDING;
		nbuildings_to_build_ = nbuildings;
	}
}

void player::select_building(int nbuilding, const boost::array<int, 4>& payment)
{
	for(int n = 0; n != 3; ++n) {
		resources_[n] -= payment[n];
		if(resources_[n] < 0) {
			tribute_[n] += resources_[n];
			resources_[n] = 0;
			ASSERT_GE(tribute_[n], 0);
		}
	}

	victory_points_ -= payment[3];

	build_building(nbuilding);

	finish_turn_(current_deity_);
}

void player::build_building(int nbuilding)
{
	buildings_.insert(nbuilding);
	--nbuildings_to_build_;
	if(nbuildings_to_build_ == 0) {
		if(ndeities_choose_) {
			state_ = STATE_CHOOSE_OTHER_DEITIES;
		} else {
			state_ = STATE_NONE;
		}
	}

	victory_points_ += building::get_building(nbuilding)->victory_points();

	building::get_building(nbuilding)->build(*this);
}

void player::select_number_of_wars(int nwars)
{

	develop(DEVELOP_MILITARY, nwars_ - nwars);
	nwars_ = nwars;
	state_ = STATE_NONE;
	finish_turn_(current_deity_);
}

void player::declare_war(boost::function<void(int)> create_war_fn)
{
	if(is_local_ai()) {
		--nwars_;
		int min_strength = 100;
		const_player_ptr target;
		foreach(const_player_ptr p, game_.players()) {
			if(p.get() != this && p->combat_strength() < min_strength) {
				target = p;
				min_strength = p->combat_strength();
			}
		}

		ASSERT_LOG(target, "Could not find target to attack");
		create_war_fn(target->number());
		return;
	}

	finish_turn_ = create_war_fn;
	state_ = STATE_DECLARING_WAR;
}

void player::make_war_declaration(int nenemy)
{
	if(nenemy >= number()) {
		++nenemy;
	}

	state_ = STATE_NONE;
	--nwars_;
	finish_turn_(nenemy);
}

int player::combat_strength() const
{
	return development_[DEVELOP_MILITARY] + military_bonus_;
}

void player::choose_tribute(player_ptr enemy, int tribute)
{
	while(tribute && std::accumulate(enemy->resources_, enemy->resources_+3, 0)) {
		const int index = rng::generate()%3;
		if(enemy->resources_[index]) {
			get_tribute(enemy, index);
			--tribute;
		}
	}
}

void player::get_tribute(player_ptr enemy, int resource)
{
	if(enemy->resources_[resource]) {
		enemy->resources_[resource]--;
	}

	tribute_[resource]++;
}

void player::choose_wars(int nwars)
{
	if(is_local_ai()) {
		nwars_ = nwars;
		int biggest_advantage = 0, biggest_deficit = 0;
		foreach(const_player_ptr p, game_.players()) {
			if(p.get() == this) {
				continue;
			}

			const int advantage = combat_strength() - p->combat_strength();
			if(advantage > biggest_advantage) {
				biggest_advantage = advantage;
			}

			if(advantage < biggest_deficit) {
				biggest_deficit = advantage;
			}
		}

		if(biggest_deficit <= -1 || biggest_advantage < 2) {
			develop(DEVELOP_MILITARY, nwars_);
			nwars_ = 0;
		}

		state_ = STATE_NONE;
		return;
	}

	state_ = STATE_CHOOSE_WAR;
	nwars_ = nwars;
}

int player::can_afford_building(int nbuilding, std::vector<boost::array<int, 4> >* solution) const
{
	const_building_ptr b = building::get_building(nbuilding);
	if(development_[DEVELOP_CULTURE] + building_culture_reduction_ < b->culture()-1) {
		return 0;
	}

	if(b->unique() && b->ninstances() > 0) {
		return 0;
	}

	int nsolutions = 0;
	int payment[4] = {0,0,0,0};

	int resources[3];
	for(int n = 0; n != 3; ++n) {
		resources[n] = resources_[n];
		if(can_use_tribute_) {
			resources[n] += tribute_[n];
		}
	}

	for(int a = 0; a <= resources[0]; ++a) {
		payment[0] = a;
		for(int b = 0; b <= resources[1]; ++b) {
			payment[1] = b;
			for(int c = 0; c <= resources[2]; ++c) {
				payment[2] = c;
				for(int d = 0; d <= max_victory_points_payment_for_building_ && d <= victory_points_; ++d) {
					payment[3] = d;
					if(is_valid_payment_for_building(nbuilding, payment)) {
						if(solution) {
							boost::array<int, 4> a;
							for(int n = 0; n != 4; ++n) {
								a[n] = payment[n];
							}

							solution->push_back(a);
						}

						++nsolutions;
					}
				}
			}
		}
	}
	return nsolutions;
}

bool player::is_valid_payment_for_building(int nbuilding, const int* resources_vec) const
{
	const_building_ptr b = building::get_building(nbuilding);
	int cost[4];
	int payment[4];
	for(int n = 0; n != 4; ++n) {
		cost[n] = b->cost()[n];
		payment[n] = resources_vec[n];
	}

	payment[3] += free_resources_toward_building_;

	for(int n = 0; n != 3; ++n) {
		if(can_use_resource_as_wildcard(n)) {
			payment[3] += payment[n];
			payment[n] = 0;
		}
	}

	for(int n = 0; n != 3; ++n) {
		payment[n] -= cost[n];
		if(payment[n] < 0) {
			payment[3] += payment[n];
			payment[n] = 0;
			if(payment[3] < 0) {
				return false;
			}
		}

		cost[n] = 0;
	}

	if(std::accumulate(payment, payment + 4, 0) != cost[3]) {
		return false;
	}


	return true;
}

bool player::has_building(int nbuilding) const
{
	return buildings_.count(nbuilding) != 0;
}

bool player::has_temple(int ndeity) const
{
	return temples_.count(ndeity);
}

void player::set_temple(int ndeity)
{
	temples_.insert(ndeity);
}

bool player::can_use_resource_as_wildcard(int nresource) const
{
	return can_use_resource_as_wildcard_[nresource];
}

void player::set_resource_as_wildcard(int nresource)
{
	can_use_resource_as_wildcard_[nresource] = true;
}

bool player::active_player_turn(boost::function<void(int)> choose_deity)
{
	state_ = STATE_CHOOSE_DEITY;
	finish_turn_ = choose_deity;

	if(is_local_ai()) {
		int choice = -1;
		int best_rating = 0;
		for(int n = 0; n != deity::num_deities(); ++n) {
			deity_ptr d = deity::get(n);
			if(d->priests().empty() == false) {
				continue;
			}

			const int rating = d->ai_rating(*this, 2);
			if(choice == -1 || rating > best_rating) {
				best_rating = rating;
				choice = n;
			}
		}

		current_deity_ = choice;

		deity_ptr d = deity::get(choice);
		move_priest_to_deity(choice);
		get_deity_benefit(choice, 2);
		return false;
	}

	return true;
}

bool player::secondary_player_turn(int ndeity, boost::function<void(int)> finish_choice)
{
	std::cerr << "PLAYER: " << number() << " SECONDARY PLAYER: " << ndeity << " " << npriests_ << "\n";
	current_deity_ = ndeity;
	finish_turn_ = finish_choice;
	state_ = STATE_CHOOSE_SECONDARY;

	deity_ptr d = deity::get(ndeity);
	if(npriests_ == 0 || d->priests().size() == 4) {
		state_ = STATE_NONE;
		finish_turn_(ndeity);
		return false;
	} else if(is_local_ai()) {
		const int rating = d->ai_rating(*this, 1);
		int nbetter = 0;
		for(int n = 0; n != deity::num_deities(); ++n) {
			deity_ptr d = deity::get(n);
			if(d->priests().empty() == false) {
				continue;
			}

			const int new_rating = d->ai_rating(*this, 2);
			if(new_rating > rating) {
				++nbetter;
			}
		}

		if(nbetter < npriests_) {
			move_priest_to_deity(ndeity);
			victory_points_ += victory_points_for_beta_;
			get_deity_benefit(ndeity, 1);
		} else {
			state_ = STATE_NONE;
			finish_turn_(current_deity_);
		}
		return false;
	}

	return true;
}

void player::collect_tribute()
{
	for(int n = 0; n != NUM_RESOURCES; ++n) {
		resources_[n] += tribute_[n];
		tribute_[n] = 0;
	}
}

void player::suffer_plague()
{
	development_[DEVELOP_POPULATION] -= (development_[DEVELOP_POPULATION]+1)/3;
}

void player::apply_population_limit()
{
	for(int n = 0; n != NUM_DEVELOPMENTS; ++n) {
		if(n == DEVELOP_CULTURE) {
			continue;
		}

		if(development_[n] > development_[DEVELOP_POPULATION]) {
			development_[n] = development_[DEVELOP_POPULATION];
		}
	}
}

void player::move_priest_to_deity(int ndeity)
{
	deity_ptr d = deity::get(ndeity);
	d->add_priest(number());
	--npriests_;

	priests_in_transit_.push_back(PriestInTransit());
	PriestInTransit& p = priests_in_transit_.back();
	p.deity_dest = ndeity;
	p.deity_pos = d->priests().size() - 1;
	p.source_pos = npriests_;
	p.time = 0;
}

bool player::priest_in_transit_to(int ndeity) const
{
	foreach(const PriestInTransit& p, priests_in_transit_) {
		if(p.deity_dest == ndeity) {
			return true;
		}
	}

	return false;
}

void player::choose_resource(int n)
{
	ASSERT_GT(nchoose_resources_, 0);
	--nchoose_resources_;
	++resources_[n];
}

void player::choose_development(int n)
{
	ASSERT_GT(nchoose_developments_, 0);
	--nchoose_developments_;
	develop(static_cast<DEVELOPMENT_TYPE>(n), 1);
}

void player::choose_regression(int n)
{
	ASSERT_GT(nchoose_regressions_, 0);
	if(development_[n] > 0 || std::accumulate(development_, development_+6, 0) == 0) {
		--nchoose_developments_;
		--development_[n];
	}
}

void player::win_war()
{
	if(gets_priest_when_win_war_) {
		++npriests_;
	}
}

void player::exchange_resources(int resource_type)
{
	resources_[resource_type]--;
	for(int n = 0; n != 3; ++n) {
		if(n != resource_type) {
			resources_[n]++;
		}
	}
}

void player::exchange_resource_for_vp(int resource_type)
{
	resources_[resource_type]--;
	victory_points_++;
}

void player::exchange_resource_for_development(RESOURCE_TYPE resource, DEVELOPMENT_TYPE development_area)
{
	resources_[resource]--;
	develop(development_area, 1);
}

void player::process()
{
	for(int n = 0; n != NUM_DEVELOPMENTS; ++n) {
		const int target = development_[n]*100;
		if(development_draw_[n] < target) {
			development_draw_[n] += 5;
		} else if(development_draw_[n] > target) {
			development_draw_[n] -= 5;
		}
	}

	foreach(PriestInTransit& p, priests_in_transit_) {
		p.time += 2;
	}

	while(priests_in_transit_.empty() == false && priests_in_transit_.front().time >= 100) {
		priests_in_transit_.erase(priests_in_transit_.begin());
	}
}

int player::ai_rate_victory_points(int n) const
{
	return n*100;
}

int player::ai_rate_development(int ntype, int num) const
{
	const DEVELOPMENT_TYPE type = static_cast<DEVELOPMENT_TYPE>(ntype);
	const int headroom = max_building_development(type) - development(type);
	if(headroom < num) {
		num = headroom;
	}

	if(type != DEVELOP_CULTURE && type != DEVELOP_POPULATION) {
		const int headroom = development(DEVELOP_POPULATION) - development(type);
		if(headroom < num) {
			num = headroom;
		}
	}

	return num*100;
}

int player::ai_rate_production(int type) const
{
	int nproductivity = resource_productivity(development(static_cast<DEVELOPMENT_TYPE>(type)));
	const int have = total_num_resources();

	const int headroom = 5 - have;
	if(nproductivity > headroom) {
		nproductivity = headroom;
	}

	return nproductivity*80;
}

int player::ai_rate_war() const
{
	int best = -1;
	foreach(const_player_ptr p, game_.players()) {
		if(p.get() != this) {
			const int diff = combat_strength() - p->combat_strength();
			if(diff > best) {
				best = diff;
			}
		}
	}

	return best*50;
}

int player::ai_rate_unleash_plague() const
{
	int damage = 0;
	const deity_ptr apollo = deity::get_by_name("apollo");
	foreach(const_player_ptr p, game_.players()) {
		if(p.get() == this) {
			continue;
		}

		if(p->has_temple(apollo->number())) {
			//they are immune, so don't bother.
			continue;
		}

		const int population = p->development(DEVELOP_POPULATION);
		const int loss = (population+1)/3;
		const int new_population = population - loss;
		for(int n = 0; n != 6; ++n) {
			if(n != DEVELOP_POPULATION && n != DEVELOP_CULTURE) {
				const int loss = p->development(static_cast<DEVELOPMENT_TYPE>(n)) - new_population;
				if(loss > 0) {
					damage += loss*30;
				}
			}
		}

		damage += loss*40;
	}

	return damage;
}

int player::ai_rate_plague_protection() const
{
	const int loss = (development(DEVELOP_POPULATION)+1)/3;
	int damage = loss*100;
	const int new_population = development(DEVELOP_POPULATION)-loss;
	for(int n = 0; n != 6; ++n) {
		if(n != DEVELOP_POPULATION && n != DEVELOP_CULTURE) {
			const int loss = development(static_cast<DEVELOPMENT_TYPE>(n)) - new_population;
			damage += loss*70;
		}
	}

	return damage;
}

