#ifndef PLAYER_HPP_INCLUDED
#define PLAYER_HPP_INCLUDED

#include <SDL.h>

#include <set>
#include <vector>

#include <boost/array.hpp>
#include <boost/function.hpp>
#include <boost/shared_ptr.hpp>

#include "building.hpp"
#include "texture.hpp"

class game;
class player;

typedef boost::shared_ptr<player> player_ptr;
typedef boost::shared_ptr<const player> const_player_ptr;

enum DEVELOPMENT_TYPE { DEVELOP_GRAIN, DEVELOP_VENISON, DEVELOP_FISH,
		                DEVELOP_CULTURE, DEVELOP_POPULATION, DEVELOP_MILITARY,
						NUM_DEVELOPMENTS };

enum PLAYER_TYPE { PLAYER_LOCAL_HUMAN, PLAYER_LOCAL_AI, PLAYER_NETWORK_HUMAN };
class player
{
public:
	static int glory_owner(int n);
	static graphics::texture victory_points_icon();
	static graphics::texture wildcard_resource_icon();
	static graphics::texture development_icon(DEVELOPMENT_TYPE type);
	static const char* get_name(int nplayer);
	static const SDL_Color& get_color(int nplayer);
	static int resource_productivity(int devel);
	player(game& g, PLAYER_TYPE type, int num);

	game& get_game() { return game_; }
	const game& get_game() const { return game_; }

	bool is_local_human() const { return type_ == PLAYER_LOCAL_HUMAN || type_ == PLAYER_NETWORK_HUMAN; }
	bool is_local_ai() const { return type_ == PLAYER_LOCAL_AI; }
	bool is_network() const { return type_ == PLAYER_NETWORK_HUMAN; }

	int number() const { return number_; }

	const int& victory_points() const { return victory_points_; }
	void get_victory_points(int points) { victory_points_ += points; }

	int max_building_development(DEVELOPMENT_TYPE type) const;
	int max_development(DEVELOPMENT_TYPE type) const;

	bool choosing_deity() const { return state_ == STATE_CHOOSE_DEITY || state_ == STATE_CHOOSE_SECONDARY; }

	void develop(DEVELOPMENT_TYPE type, int amount);
	void produce(RESOURCE_TYPE type);

	void develop_or_produce(RESOURCE_TYPE type, int num=1);
	bool choosing_develop_or_produce() const { return state_ == STATE_DEVELOP_OR_PRODUCE; }
	void choose_develop();
	void choose_produce();

	int choosing_resource() const { return state_ == STATE_PRODUCE_CHOICE ? develop_or_produce_num_choices_ : 0; }
	int choosing_development() const { return state_ == STATE_DEVELOP_CHOICE ? develop_or_produce_num_choices_ : 0; }

	void choose_resources_to_develop_or_produce(RESOURCE_TYPE resource);

	int development_draw(DEVELOPMENT_TYPE type) const { return development_draw_[type]; }
	int development(DEVELOPMENT_TYPE type) const { return development_[type]; }
	int resources(RESOURCE_TYPE type) const { return resources_[type]; }
	int tribute(RESOURCE_TYPE type) const { return tribute_[type]; }

	int total_num_resources() const;

	void get_resource(int type, int amount=1) { resources_[type] += amount; }

	void suffer_plague();
	void apply_population_limit();
	void restore_priests();
	void get_priest_back() { ++npriests_; }
	int max_priests() const;
	int num_priests() const { return npriests_; }

	void select_deity(int n);
	void select_pass();

	void choose_other_deities(int n);
	int choosing_other_deities() const;

	void choose_plague();
	void select_plague(bool plague);
	bool choosing_plague() const { return state_ == STATE_CHOOSE_PLAGUE; }

	void choose_building(int nbuildings);
	void select_building(int nbuilding, const boost::array<int, 4>& payment);
	bool choosing_building() const { return state_ == STATE_CHOOSE_BUILDING; }

	void build_building(int nbuilding);

	void choose_wars(int nwars);
	int choosing_number_of_wars() const { return state_ == STATE_CHOOSE_WAR ? nwars_ : 0; }
	void select_number_of_wars(int nwars);

	int number_of_wars() const { return nwars_; }

	bool declaring_war() const { return state_ == STATE_DECLARING_WAR; }

	void declare_war(boost::function<void(int)> create_war_fn);

	void make_war_declaration(int nenemy);

	int can_afford_building(int nbuilding, std::vector<boost::array<int, 4> >* solution=0) const;
	bool is_valid_payment_for_building(int nbuilding, const int* resources_vec) const;

	bool has_building(int nbuilding) const;

	bool has_temple(int ndeity) const;
	void set_temple(int ndeity);

	bool can_use_resource_as_wildcard(int nresource) const;
	void set_resource_as_wildcard(int nresource);

	void get_deity_benefit(int ndeity, int power);

	bool active_player_turn(boost::function<void(int)> choose_deity);
	bool secondary_player_turn(int ndeity, boost::function<void(int)> finish_choice);

	void collect_tribute();

	const std::vector<int>& deities_unavailable() const { return deities_unavailable_; }

	void set_max_culture_level(int value) { max_culture_level_ = value; }
	void set_max_population_level(int value) { max_population_level_ = value; }
	void set_max_military_level(int value) { max_military_level_ = value; }
	void set_building_culture_reduction(int value) { building_culture_reduction_ = value; }
	void set_production_slider_increase(int value) { production_slider_increase_ = value; }

	void set_population_priest_requirement(int value) { population_priest_requirement_ = value; }
	void set_culture_priest_requirement(int value) { culture_priest_requirement_ = value; }

	void add_military_bonus(int value) { military_bonus_ += 1; }
	void set_military_win_tie() { military_win_tie_ = true; }

	int combat_strength() const;
	bool military_win_tie() const { return military_win_tie_; }

	void choose_tribute(player_ptr enemy, int tribute);
	void get_tribute(player_ptr enemy, int resource);

	void add_victory_points_for_beta(int value) { victory_points_for_beta_ += value; }

	bool priest_in_transit_to(int ndeity) const;

	struct PriestInTransit {
		int deity_dest;
		int deity_pos;
		int source_pos;
		int time;
	};
	const std::vector<PriestInTransit>& priests_in_transit() const { return priests_in_transit_; }

	int number_of_resources_choosing() const { return nchoose_resources_; }
	void add_free_resource_choice(int n) { nchoose_resources_ += n; }
	void choose_resource(int n);

	int number_of_developments_choosing() const { return nchoose_developments_; }
	void add_free_development_choice(int n) { nchoose_developments_ += n; }
	void choose_development(int n);

	int number_of_regressions_choosing() const { return nchoose_regressions_; }
	void add_free_regression_choice(int n) { nchoose_regressions_ += n; }
	void choose_regression(int n);

	void set_gets_priest_when_win_war() { gets_priest_when_win_war_ = true; }
	void win_war();

	void allow_victory_points_payment_for_building(int amount) { max_victory_points_payment_for_building_ += amount; }

	bool can_exchange_resources() const { return can_exchange_resources_; }
	void set_can_exchange_resources() { can_exchange_resources_ = true; }

	void exchange_resources(int resource_type);

	bool can_exchange_resource_for_vp() const { return can_exchange_resource_for_vp_; }
	void set_can_exchange_resource_for_vp() { can_exchange_resource_for_vp_ = true; }
	void exchange_resource_for_vp(int resource_type);

	void add_free_resource_toward_building() { free_resources_toward_building_++; }

	bool can_exchange_resource_for_development() { return can_exchange_resource_for_development_; }
	void set_can_exchange_resource_for_development() { can_exchange_resource_for_development_ = true; }

	void exchange_resource_for_development(RESOURCE_TYPE resource, DEVELOPMENT_TYPE develop);

	void set_can_use_tribute() { can_use_tribute_ = true; }

	bool can_take_tribute_from_reserve() const { return can_take_tribute_from_reserve_; }
	void set_can_take_tribute_from_reserve() { can_take_tribute_from_reserve_ = true; }

	bool get_vp_for_priests_on_beta() const { return get_vp_for_priests_on_beta_; }
	void set_vp_for_priests_on_beta() { get_vp_for_priests_on_beta_ = true; }
	bool get_development_for_priests_on_beta() const { return get_development_for_priests_on_beta_; }
	void set_development_for_priests_on_beta() { get_development_for_priests_on_beta_ = true; }

	void process();

	int ai_rate_victory_points(int n) const;
	int ai_rate_development(int type, int num=1) const;
	int ai_rate_production(int type) const;
	int ai_rate_war() const;
	int ai_rate_unleash_plague() const;
	int ai_rate_plague_protection() const;

private:
	game& game_;

	PLAYER_TYPE type_;
	int development_[NUM_DEVELOPMENTS];
	int resources_[NUM_RESOURCES];
	int tribute_[NUM_RESOURCES];

	bool can_use_resource_as_wildcard_[NUM_RESOURCES];

	//the positions we draw the development slider items in. This is
	//expressed as 100x the items in development_ and adjusts progressively
	//to match reality to create a sliding effect.
	int development_draw_[NUM_DEVELOPMENTS];

	int victory_points_;

	int max_victory_points_payment_for_building_;

	std::set<int> buildings_;
	std::set<int> temples_;
	int number_;

	int npriests_;
	int current_deity_;

	std::vector<PriestInTransit> priests_in_transit_;
	void move_priest_to_deity(int ndeity);

	boost::function<void(int)> finish_turn_;

	enum PLAYER_STATE { STATE_NONE, STATE_CHOOSE_DEITY, STATE_CHOOSE_SECONDARY, STATE_DEVELOP_OR_PRODUCE, STATE_CHOOSE_OTHER_DEITIES, STATE_DEVELOP_CHOICE, STATE_PRODUCE_CHOICE, STATE_CHOOSE_PLAGUE, STATE_CHOOSE_BUILDING, STATE_CHOOSE_WAR, STATE_DECLARING_WAR };
	PLAYER_STATE state_;

	//valid if state_ == STATE_DEVELOP_OR_PRODUCE
	RESOURCE_TYPE resource_type_choosing_;
	int develop_or_produce_num_choices_;
	RESOURCE_TYPE develop_or_produce_choice_excluded_;

	//valid if state_ == STATE_CHOOSE_OTHER_DEITIES
	std::vector<int> deities_unavailable_;
	int ndeities_choose_;

	//valid if state_ == STATE_CHOOSE_BUILDING
	int nbuildings_to_build_;

	//valid if state_ == STATE_CHOOSE_WAR
	int nwars_;

	int max_culture_level_, max_population_level_, max_military_level_;
	int building_culture_reduction_;
	int production_slider_increase_;

	int population_priest_requirement_, culture_priest_requirement_;

	int military_bonus_;
	bool military_win_tie_;

	//number of victory points you get for sending a priest to a beta position.
	int victory_points_for_beta_;

	int nchoose_resources_, nchoose_developments_, nchoose_regressions_;

	bool gets_priest_when_win_war_;

	//in the upkeep phase allows exchanging a resource for the other two.
	bool can_exchange_resources_;

	//in the upkeep phase allows exchanging a resource for a VP.
	bool can_exchange_resource_for_vp_;

	bool can_exchange_resource_for_development_;

	int free_resources_toward_building_;

	bool can_use_tribute_;

	bool can_take_tribute_from_reserve_;

	bool get_vp_for_priests_on_beta_, get_development_for_priests_on_beta_;
};

#endif
