#include <algorithm>
#include <iostream>
#include <string>

#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>

#include "asserts.hpp"
#include "foreach.hpp"
#include "server.hpp"

using boost::asio::ip::tcp;

server::server(boost::asio::io_service& io_service)
  : acceptor_(io_service, tcp::endpoint(tcp::v4(), 17000))
{
	start_accept();
}

void server::start_accept()
{
	socket_ptr socket(new tcp::socket(acceptor_.io_service()));
	acceptor_.async_accept(*socket, boost::bind(&server::handle_accept, this, socket, boost::asio::placeholders::error));
}

void server::handle_accept(socket_ptr socket, const boost::system::error_code& error)
{
	if(error) {
		std::cerr << "ERROR IN ACCEPT\n";
		return;
	}

	std::cerr << "RECEIVED CONNECTION. Listening\n";

	if(!game_starting_) {
		game_starting_.reset(new game);
	}

	game_starting_->players.push_back(socket);

	client_info& info = clients_[socket];
	info.game = game_starting_;

	start_receive(socket);

	start_accept();
}

void server::start_receive(socket_ptr socket)
{
	buffer_ptr buf(new boost::array<char, 1024>);
	socket->async_read_some(boost::asio::buffer(*buf), boost::bind(&server::handle_receive, this, socket, buf, _1, _2));
}

void server::handle_receive(socket_ptr socket, buffer_ptr buf, const boost::system::error_code& e, size_t nbytes)
{
	if(e) {
		disconnect(socket);
		return;
	}

	handle_incoming_data(socket, &(*buf)[0], &(*buf)[0] + nbytes);

	start_receive(socket);
}

void server::handle_incoming_data(socket_ptr socket, const char* i1, const char* i2)
{
	client_info& info = clients_[socket];
	const char* string_terminator = std::find(i1, i2, 0);
	info.partial_message.insert(info.partial_message.end(), i1, string_terminator);
	if(string_terminator != i2) {
		info.partial_message.push_back(0);
		handle_message(socket, info.partial_message);
		info.partial_message.clear();
		handle_incoming_data(socket, string_terminator+1, i2);
	}
}

void server::handle_message(socket_ptr socket, const std::vector<char>& msg)
{
	try {
		std::cerr << "MSG: " << &msg[0] << "\n";
		simple_wml::document doc(&msg[0], simple_wml::INIT_STATIC);
		handle_message(socket, doc);
	} catch(simple_wml::error& e) {
		std::cerr << "INVALID WML RECEIVED\n";
	}
}

void server::handle_message(socket_ptr socket, simple_wml::document& msg)
{
	const simple_wml::node& node = msg.root();
	const simple_wml::string_span& name = node.first_child();
	const simple_wml::node* first_node = node.first_child_node();
	client_info& info = clients_[socket];

	if(name == "update_game") {
		std::ostringstream s;
		s << "[game]\nplayers=\"" << (game_starting_ ? game_starting_->players.size() : 0) << "\"\n[/game]\n";
		send_msg(socket, s.str());
	} else if(name == "start_game") {
		if(game_starting_) {
			for(int n = 0; n != game_starting_->players.size(); ++n) {
				std::ostringstream s;
				s << "[start_game]\nplayers=\""
				  << game_starting_->players.size() << "\"\n"
				  << "position=\"" << n << "\"\n"
				  << "[/start_game]\n";
				send_msg(game_starting_->players[n], s.str());
			}
		}

		game_starting_.reset();
	} else {
		if(info.game) {
			foreach(socket_ptr sock, info.game->players) {
				if(sock == socket) {
					continue;
				}

				send_msg(sock, msg.output());
			}
		}
	}
}

void server::send_msg(socket_ptr socket, const std::string& msg)
{
	std::vector<char> buf(msg.begin(), msg.end());
	buf.push_back(0);
	boost::asio::async_write(*socket, boost::asio::buffer(buf),
			                         boost::bind(&server::handle_send, this, socket, _1, _2));
	std::cerr << "SEND MSG: " << msg.c_str() << "\n";
}

void server::handle_send(socket_ptr socket, const boost::system::error_code& e, size_t nbytes)
{
	if(e) {
		disconnect(socket);
	}
}

void server::disconnect(socket_ptr socket)
{
	client_info& info = clients_[socket];
	if(info.game) {
		info.game->players.erase(std::remove(info.game->players.begin(), info.game->players.end(), socket), info.game->players.end());
	}
	clients_.erase(socket);
}

int main()
{
	boost::asio::io_service io_service;
	server s(io_service);
	try {
		io_service.run();
	} catch(...) {
		std::cerr << "ERROR CAUGHT. EXITING\n";
	}
}
